package net.pricefx.formulaengine;

import java.util.Collection;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import net.pricefx.server.dto.calculation.ComplexCalculationResult;
import net.pricefx.server.util.metamodel.Column;
import org.joda.time.DateTime;

import com.googlecode.genericdao.search.Filter;

import net.pricefx.common.util.CalendarUtil.TimePeriod;
import net.pricefx.common.util.CalendarUtil.TimeUnit;
import net.pricefx.domain.CustomerGroup;
//import net.pricefx.domain.DMCalendar;
import net.pricefx.domain.ProductGroup;
import net.pricefx.formulaengine.scripting.Matrix2D;
import net.pricefx.server.dto.calculation.ResultMatrix;
//import net.pricefx.server.dto.calculation.ResultPAQuery;
import net.pricefx.server.protocol.StreamResults;
//import net.pricefx.server.util.metamodel.dm.Column;
//import net.pricefx.server.util.metamodel.dm.Table;


/**
 * Provides an API to query and load PA data from a formula context.
 *
 * <p><b>Example:</b></p>
 * <p>This example demonstrates how to run a query to read data from PriceAnalyzer datamart and return it as in a form of {@link ResultMatrix} </p>
 * <blockquote>
 *     <pre>
 *                     def ctx = api.getDatamartContext()
 *                     def dm = ctx.getDatamart("txDM")
 *                     def query = ctx.newQuery(dm)
 *                     query.select("invoiceDateQuarter", "Q")
 *                     query.select("SUM(amount)", "R")
 * 					   query.orderBy("Q")
 *                     return ctx.executeQuery(query)?.data?.toResultMatrix().entries
 *     </pre>
 * </blockquote>
 *
 * @see DatamartContext#newQuery(Table)
 * @see DatamartContext#executeQuery(Query)
 * @see ResultMatrix
 */
public interface DatamartContext {

	/**
	 * <p>Helper class mainly for working with dates and time periods as used in the PriceAnalyzer module.  In PA, for each Date and DateTime field defined  
	 * as a dimension, the application generates derived fields to denote the calendar 'Week', 'Month', 'Quarter' and 'Year'.  For example, a PricingDate = '2015-01-02'
	 * has a corresponding: 
	 * <ul>
	 *   <li>PricingDateWeek = '2015-W01'</li>
	 *   <li>PricingDateMonth = '2015-M01'</li>
	 *   <li>PricingDateQuarter = '2015-Q1'</li>
	 *   <li>PricingDateYear = '2015'</li>
	 * </ul>
	 * 
	 * A DateTime field gets an additional 'Day' field.
	 * 
	 * <p>DM queries can specify a filter on these calendar fields and more importantly, roll up data along them.
	 * The calendar field values are further abstracted in a TimePeriod.  A TimePeriod has a start period and an end period (singular TimePeriods have start = end),
	 * and a TimeUnit. A TimePeriod also has a start date and an end date which define its alignment with the commonly known {@link java.util.Calendar}.  
	 *  
	 * <p>A TimeUnit can be:
	 * <ul>
	 *   <li>DAY, named 'Day'</li>
	 *   <li>WEEK, named 'Week'</li> 
	 *   <li>MONTH, named 'Month'</li>
	 *   <li>QUARTER, named 'Quarter'</li>
	 *   <li>YEAR, named 'Year'</li>
	 * </ul>
	 * The name can be used in various helper methods accepting a string argument rather than the enumeration value.
	 */
	public interface Calendar {
		/**
		 * Helper method to build a filter representing <code>dateDimFieldName</code> in <code>[timePeriod.startPeriod, timePeriod.endPeriod]</code>
		 * Note the inclusion on both ends; this is to accommodate the most frequent use case where the TimePeriod is a single period, for
		 * example InvoiceDate in [2015,2015]. An example of a non-singular period filter is [2015-Q1, 2015-Q2]; see {@link TimePeriod}
		 * for more details.
		 *   
		 * @param timePeriod {@link TimePeriod} that the <code>dateDimFieldName</code> field needs to match.
		 * @param dateDimFieldName Datamart field of the type 'Date' to filter on.
		 * @return {@link Filter} limiting the given date field to the given time period.
		 */
		Filter timePeriodAsFilter(TimePeriod timePeriod, String dateDimFieldName);

		/**
		 * Returns the TimePeriod object matching its string representation. For example, '2015-Q1' represents a TimePeriod
		 * with the start date 2015-01-01 and end date 2015-03-31 and TimeUnit 'Quarter'. The start and end periods in this example
		 * are both equal to '2015-Q1'.  The same time span could be represented by different TimePeriods, such as 
		 * [2015-M01, 2015-M3], though the latter would have a TimeUnit equal to 'Month', and start and end periods of 2015-M01
		 * and 2015-M03. Different representations are chosen to fit different use cases. TimePeriods are typically used in 
		 * the build-up to a {@link Filter} or {@link DataSlice} definition.  
		 *  
		 * @param periodAsString Textual representation of a TimePeriod, such as '2015', '2015-Q1' etc.
		 * @return The Corresponding TimePeriod object to be used as an argument in other calendar dependent API calls.
		 */
		TimePeriod getTimePeriod(String periodAsString);
		
		/**
		 * Parses a date, as well as a named TimeUnit, resulting in a TimePeriod with that unit
		 * that encompasses the given date. For example, getTimePeriodByDate("2015-01-02", "Month") results in the [2015-M01] TimePeriod. 
		 * @param dateAsString 'yyyy-MM-dd' string representation of {@link java.util.Date}. 
		 * @param unitAsString String representation of a TimeUnit.
		 * @return Matching TimePeriod object.
		 */
		TimePeriod getTimePeriodByDate(String dateAsString, String unitAsString);	

		/**
		 * Useful for constructing Options lists (user entry / input parameters). Returns a chronologically ordered list of TimePeriod
		 * names starting with the given start period, with each TimePeriod expressed in the given TimeUnit.
		 * For example, getTimePeriodNames("2015-Q01", "2015-Q3", "Month") results in ["2015-M01", "2015-M02", ..., "2015-M09"].  
		 * @param startPeriodAsString String representation of the start TimePeriod. 
		 * @param endPeriodAsString String representation of the end TimePeriod.
		 * @param unitAsString String representation of a TimeUnit.
		 * @return
		 */
		List<String> getTimePeriodNames(String startPeriodAsString, String endPeriodAsString, String unitAsString);

		/**
		 * The same as {@link getTimePeriodNames} but it returns a list of TimePeriod objects, rather than the corresponding period names.
		 * @param startPeriodAsString
		 * @param endPeriodAsString
		 * @param unitAsString
		 * @return
		 */
		List<TimePeriod> getTimePeriods(String startPeriodAsString, String endPeriodAsString, String unitAsString);

		List<TimePeriod> getTimePeriods(Date startDate, Date endDate, TimeUnit unit);

		/**
		 * Returns the start date of a given TimePeriod. 
		 * @param period String representation of a TimePeriod.
		 * @return Period's start Date.
		 */
		Date getStartDate(String period);
		
		/**
		 * Returns the end date of a given TimePeriod. 
		 *
		 * @param period String representation of a TimePeriod.
		 * @return Period's end Date.
		 */
		Date getEndDate(String period);

		/**
		 * Parses a period string and offsets the start date of the resulting TimePeriod with a number of TimeUnits.
		 * <p>Examples:
		 * <ul>  
		 * 		<li> add("2015", 3, "Month") = "2015" </li>
		 * 		<li> add("2015", -3, "Month") = "2014" </li>
		 * 		<li> add("2015-M01", 12, "Month") = "2014-M01" </li>
		 * </ul></p>
		 * 
		 * @param periodAsString String representation of the end TimePeriod.
		 * @param numberOfUnits Number of TimeUnits to use as the offset.
		 * @param unitAsString String representation of a TimeUnit.
		 * @return Resulting offset TimePeriod in the same TimeUnit as the input period.
		 */
		String add(String periodAsString, int numberOfUnits, String unitAsString);

		/**
		 * Calculates the number of days between the start date of period1 and end date of period2 (both inclusive).
		 * For example, getDaysDiff("2015-M04", "2015-M05") = 61, since there are 30 days in April and 31 days in May.
		 * @param endPeriod String representation of the end TimePeriod.
		 * @param startPeriod String representation of the start TimePeriod.
		 * @return Number of days covering [startPeriod.startDate, endPeriod.endDate].
		 */
		long getDaysDiff(String endPeriod, String startPeriod);

		/**
		 * Not so much a PA specific helper method: best attempt at converting a string representation (not necessarily 
		 * in the 'yyyy-MM-dd' format) to a Date.
		 * @param dateAsString String representation of a Date.
		 * @return Matching Date or <code>null</code> if no match.
		 */
		Date parseDate(String dateAsString);
		
		/**
		 * Not so much a PA specific helper method: best attempt at converting a string representation (not necessarily 
		 * in the 'yyyy-MM-dd'T'HHmmss' format) to a DateTime.
		 * @param datetimeAsString String representation of a DateTime.
		 * @return Matching {@link org.joda.time.DateTime} or <code>null</code> if no match.
		 */
		DateTime parseDateTime(String datetimeAsString);
	}
	
	
	
	
	/**
	 * Instantiates a PA Calendar utility object.
	 * @return {@link net.pricefx.common.util.CalendarUtil Calendar} 
	 */
	 public Object calendar();
	 
	/**
	 * Gets a table object representing a DataFeed with the given name. A reference to this table can be used when building
	 * a {@link Query} on DataFeedLoad.
	 * @param name sourceName, uniqueName or label of the DF.
	 * @return Table representing the DF in DataContext. 
	 * */
	public Object getDataFeed(String name);

	/**
	 * Gets a table object representing a DataSource with the given name. A reference to this table can be used when building
	 * a {@link Query} on that DataSource.
	 * @param name sourceName, uniqueName or label of the DS.
	 * @return Table representing the DS in DataContext.
	 */
	public Object getDataSource(String name);

	/**
	 * Gets a table object representing a Datamart with the given name. A reference to this table can be used when building
	 * a {@link Query} on that Datamart.
	 * @param name sourceName, uniqueName or label of the DM.
	 * @return Table representing the DM in DataContext.
	 * */
	public Object getDatamart(String name);

	/**
	 * Gets a table object representing a Model with the given name. A reference to this table can be used when building
	 * a {@link Query} on that Model.
	 * @param name sourceName, uniqueName or label of the Model.
	 * @return Table representing the Model in DataContext.
	 * */
	public Object getModel(String name);

	/**
	 * Get a table object representing the FC with the given source name.  A source name consists of two parts, separated with a '.':
	 * <p><ul>
	 * 		<li> 'DMF', 'DMDS', 'DM' or 'DMSIM', i.e. the typeCode of DataFeed, DataSource, Datamart and Sim-Datamart resp. </li>
	 * 		<li> the FC's uniqueName </li>
	 * </ul></p>
	 * A reference to this tbale can then be used when building a {@link Query} on that FC.
	 * 
	 * @param sourceName The sourceName of the FC.
	 * @return Table representing the FC in the DataContext.
	 */
	public Object getFieldCollection(String sourceName);

	/**
	 * Gets a table object representing a FieldCollection representing a rollup with the given label. If more than one such rollup exists, the first
	 * one found is returned, in no particular or predefined order.
	 * @param label    Label of the Rollup.
	 * @return         Table representing the FC form of the rollup in DataContext.
	 */
    public Object getRollup(String label);

	/**
	 * @param table Table representing the FC to query.
	 * @return {@link Query} builder.
	 */
	public Query newQuery(Object table);

	/**
	 * Builds a new query on the given table. 
	 * @param table Represents the FC to query.
	 * @param rollup Line level or rollup query.
	 * @return {@link Query} builder.
	 */
	public Query newQuery(Object table, boolean rollup);

	/**
	 * Creates a new query from an existing one. This pattern can be used to build a base or template query, and then
	 * instantiate several variants, for example each for a different DataSlice (this year sales data vs. last year etc.).
	 * 
	 * @param otherQuery Query to use as a basis for a new query.
	 * @return {@link Query} builder.
	 */
	public Query newQuery(Query otherQuery);

	@Deprecated
	public DatamartContext.Query newQuery(DatamartContext.Query query1, DatamartContext.Query query2, LinkedHashMap<String,String> joinFieldsMap);
	
	@Deprecated	
	public DatamartContext.Query newQuery(DatamartContext.Query query1, DatamartContext.Query query2, LinkedHashMap<String,String> joinFieldsMap, boolean rollup);

	/**
	 * The preferred method for running a join query is now  {@link #executeSqlQuery}.
	 *  
	 * @param query1  Sub-source1 represented by the result of query1.
	 * @param query2  Sub-source2 represented by the result of query2.
	 * @param joinFieldsMap	Map of fields in source1 to fields in source1, representing the join condition.
	 * @param joinMode	How to combine source1 and source2 into one, to be the source to this query.  Note that the source field names adhere to different naming schemes depending on the join mode:
	 * <p><ul>
	 * 		<li> 'INNER':	Join semantics as in the SQL inner join. If source2 has the same named fields as source1, they are not added to the combined source.  
	 * 		<li> 'LEFT_OUTER':	As in SQL left outer join. If source2 has same named fields as source1, they are not added to the combined source.  
	 * 		<li> 'FULL_OUTER':	As in SQL full outer join. Fields in the combined source get a '_&lt;query.alias&gt;' postfix (defaults to '_1' for source1 and '_2' for source2).
	 * 		<li> 'INNER_ALL':	As in SQL inner join. Fields in the combined source get a postfix as above.  
	 * 		<li> 'LEFT_OUTER_ALL':	As in SQL left outer join. Fields in the combined source get a postfix as above.  
	 * </ul></p>
	 * @param rollup	True if this is to be a rollup query, i.e. with a Group by clause.
	 * @return	New query (select) on the combined source.
	 */
	@Deprecated
	public DatamartContext.Query newQuery(DatamartContext.Query query1, 
	        							  DatamartContext.Query query2, 
	        							  LinkedHashMap<String,String> joinFieldsMap, 
	        							  String joinMode, boolean rollup);

	/**
	 * Executes the given {@link Query}. For a rollup query, if the internal row limit (set by 'datamart.query.internalRowLimit' Pricefx instance
	 * param) is exceeded, a <code>null</code> is returned. The rationale for this behaviour is that a rollup, or so called analytical query result,
	 * is unreliable if not all data in scope could be examined. This is different to a fetch or paging query, which can safely request one page at the time.
	 * In syntax check mode, a maximum of 200 rows is returned in all cases.
     * Sample code:
     * <pre><code>
     *     def ctx = api.getDatamartContext()
     *     def dm = ctx.getDatamart("Transaction DM")
     *     def query = ctx.newQuery(dm)
     *     query.select("CustomerId")
     *     query.select("MaterialID")
     *     query.select("SUM(Sales)", "Revenue")
     *     query.select("SUM(Quantity)", "Volume")
     *     def result = ctx.executeQuery(query)  
     *     for (def r=0; r &lt; result.data.getRowCount(); r++){
     *         def row = result.data.getRowValues(r)    // row #r as map
     *         api.trace("query", "row $r", row)
     *     }    
     * </code></pre> 
	 * @param query  {@link Query} to execute.
	 * @return {@link net.pricefx.formulaengine.DatamartQueryResult DatamartQueryResult} providing data in the Matrix2D form and summary information in the map form.
	 */
    DatamartQueryResult executeQuery(Query query) throws InterruptedException;
	
    /**
     * Executes the given {@link Query} and returns a result, so that one row at the time can be examined. This is different to <code>executeQuery</code>, which 
     * always returns the full data set in the scope of the query. When a row is retrieved and moved onwards from, it is no longer available to the client code.
     * The typical usage is to consume the result row by row, processing it into some accumulating data structure.<br>
     * <b>Important note:</b> streamQuery is not executed in the syntax check mode and returns null.<br>
     * <p>Sample code:
     * <pre><code>
     *     def ctx = api.getDatamartContext()
     *     def dm = ctx.getDatamart("Transaction DM")
     *     def query = ctx.newQuery(dm)
     *     query.select("CustomerId")
     *     query.select("MaterialID")
     *     query.select("SUM(Sales)", "Revenue")
     *     query.select("SUM(Quantity)", "Volume")
     *     def results = ctx.streamQuery(query)  
     *     def r=0
     *     while(results.next()){
     *         def row = results.get()  // current row as map
     *         api.trace("streamQuery", "row $r", row)
     *         r++
     *     }
     *     results.close()
     * </code></pre> 
     * @param query  {@link Query} to execute.
     * @return StreamResults Similar to a JDBC ResultSet, but only implementing next(), get() and close().
     * @throws InterruptedException
     */
    StreamResults streamQuery(Query query) throws InterruptedException;

    /**
     * Executes an ANSI compliant SQL SELECT statement in the PA DB.<br>
     * <b>Important note:</b> A non-compliant statement that does not fail at the present time may well fail in the future releases.<br>
     * The DB schema that can be queried is constructed on the fly by means of view definitions. A view is defined using the usual query API.
     * At least one view needs to be defined. The first view gets an alias 'T1', the second 'T2' etc.<br>  
     * The columns of a view are named from the defining query projections' aliases.  Note that in the SQL standard the identifiers need to be
     * double-quoted to preserve the case. Therefore, unless a projection alias is all lowercase, the SQL statement will need to double quote references
     * to a view's column names.   
     * <p>Example:<br><pre><code>
     * 		def ctx = api.getDatamartContext()
     * 		def dm = ctx.getDatamart("TransactionsDM")
     * 		def ds = ctx.getDataSource("ForecastDS")
     * 		def t1 = ctx.newQuery(dm)
     * 		t1.select("ProductID", "product")
     * 		t1.select("SUM(InvoicePrice)", "revenue")
     * 		t1.select("SUM(Quantity)", "volume")
     * 
     * 		def t2 = ctx.newQuery(ds, false)
     * 		t1.select("ProductID", "product")
     * 		t2.select("Revenue, "revenue")
     * 		t2.select("Volume", "volume")
     * 
     * 		def sql = """ SELECT T1.product, T1.revenue AS ActualRevenue, T2.revenue AS ForecastRevenue
     * 						T1.volume AS ActualVolume, T2.volume AS ForecastVolume
     * 						FROM T1 LEFT OUTER JOIN T2 USING (product) """
     * 		return ctx.executeSqlQuery(sql, t1, t2)?.toResultMatrix()
     * </code></pre>
     * @param sources	Views that make up the DB schema that can be queried, in the form of query definitions of type {@link DatamartContext.Query}
     * 					or Strings representing 'SELECT' statements that will be added to the final SQL statement's WITH clause. The sources are assigned
     * 				    the 'Ti' relation alias in the order of appearance in the source Collection ((T1 for the first source).		
     * @return	Query result as a Matrix2D object.
     * @throws InterruptedException
     */
	Matrix2D executeSqlQuery(String sql, Object... sources) throws InterruptedException;

	/**
	 * Instantiates a query object for building SQL statements from source queries, with clauses and parameter bindings.
	 * To be executed by {@link #executeSqlQuery}.
	 * <p>Example:<br><pre>
     *     def ctx = api.getDatamartContext()
     *     def dm = ctx.getDatamart("TransactionsDM")
     *     def ds = ctx.getDataSource("ForecastDS")
     *     def t1 = ctx.newQuery(dm)
     *                 .select("ProductID", "product")
     *                 .select("ProductGroup", "PG")
     *                 .select("SUM(InvoicePrice)", "revenue")
     *                 .select("SUM(Quantity)", "volume")
     *
     *     def t2 = ctx.newQuery(ds, false)
     *                  .select("ProductID", "product")
     *                  .select("Revenue, "revenue")
     *                  .select("Volume", "volume")
     *
     *     def sqlQuery = ctx.newSqlQuery()
     *                       .addSource(t1)
     *                       .addSource(t2)
     *
     *     def with = """ SELECT T1.product, T1.revenue AS ActualRevenue, T2.revenue AS ForecastRevenue
     *                           T1.volume AS ActualVolume, T2.volume AS ForecastVolume
     *                           FROM T1 LEFT OUTER JOIN T2 USING (product)
     *                           WHERE T1.PG = ? """
     *     sqlQuery.addWith(with, "PG-ABC") 	// binding some product group value;  with-clause gets assigned the T3 alias
     *     def sql = " SELECT SUM(ActualRevenue) - SUM(ForecastRevenue) FROM T3 "
     *     sqlQuery.setQuery(sql)
     *     return ctx.executeSqlQuery(sqlQuery)?.toResultMatrix()
	 * </pre>
	 * @return New DMSqlQuery object.
	 */
	Object newSqlQuery();

	/**
	 * Executes an ANSI compliant SQL SELECT statement in the PA DB.<br>
	 * @param sqlQuery  SQL statement definition.
	 * @return Query result as a Matrix2D object.
	 * @throws InterruptedException
	 */
	Matrix2D executeSqlQuery(Object sqlQuery) throws InterruptedException;


    /**
     * <p>EXPERIMENTAL: Not all aspects and properties of a query are supported. For example, currently not supported are dim filters, row limit, join queries etc.
     * 
     * <p>Builds and validates a query to be rendered in the client. Technically, a DataTransferObject representing the query is created, compatible
     * with the query format in the REST API.
     * @param query Query to be rendered in the client.
     * @return  ResultPAQuery (calculation result) embedding a REST API compatible map representation of the query.
     */
	Object buildQuery(DatamartContext.Query query);

    /**
     * Calculates "Min", "Max", "#", "#Nulls", "#Distinct", "Sample" for dimension projections, 
     * and "Min", "Max", "Mean", "Std", "#", "#Nulls" for numeric projections.
     * @param query  Query defining the data to a profile.
     * @return  DataProfilerResult split up in dimensions and numeric projections results. 
     */
    DataProfilerResult profileData(Query query);
        
	/**
	 * Creates a new DatamartSlice which allows for setting filter criteria along the Time, CustomerGroup, ProductGroup
	 * or any other dimension in a Datamart.
	 *
	 * @return Instantiated, empty {@link DataSlice} object.
	 */
	DataSlice newDatamartSlice();

	/**
	 * Creates a new DatamartSlice which allows for setting filter criteria along the Time, CustomerGroup, ProductGroup
	 * or any other dimension in a Datamart, initialized with the name of the time dimension field and an optional set of
	 * filter criteria.
	 *
	 * @param dateFieldName Name of the time dimension field.
	 * @param timePeriodsAndProductAndCustomerGroups TimePeriod, CustomerGroup, ProductGroup filters.
	 * @return Initialized {@link DataSlice} object.
	 */
	Object newDatamartSlice(String dateFieldName, Object... timePeriodsAndProductAndCustomerGroups);


    
	/**
	 * API for building a PA query in a formula context.
	 * Note that most methods return the {@link Query} object being manipulated to allow for builder-like usage.
	 */
	interface Query {

	    Object getTable();

		/**
		 * Equivalent to the DISTINCT option in an SQL select statement.
	     * @return This {@link Query} 
		 */
	    Query selectDistinct();

		/**
		 * Adds a COUNT(*) projection to the query.
	     * @return This {@link Query} 
		 */
	    Query selectCount();

	    /**
	     * Adds a projection to the query.
	     * @param expr  Expression referencing any queryable field in the query's source FC.
	     * @param alias  Projection alias; helpful when accessing the query result data rows.
	     * @return This {@link Query} 
	     */
	    Query select(String expr, String alias);
	    
	    Query select(String expr, String alias, String... statistics);

	    Query selectAll();

	    Query selectAll(boolean fieldNameAsAlias);

	    /*
	     * Adds a condition to the query's WHERE clause.
	     * @param slice {@link DataSlice} for which the filter representation is to be added to the query's WHERE clause. 
	     * @return This {@link Query} 
	     * @throws Exception
	     */
	    Query where(DataSlice slice) throws Exception;

	    /**
	     * Adds a ProductGroup filter condition to the query's WHERE clause, for example a product group from an InputEntry("ProductGroup") parameter.
	     * @param productGroup ProductGroup for which the filter representation is to be added to the query's WHERE clause. 
	     * @return This {@link Query} 
	     * @throws Exception
	     */
	    Query where(ProductGroup productGroup) throws Exception;
	    
	    /**
	     * Adds a CustomerGroup condition to the query's WHERE clause, for example a customer group from an InputEntry("CustomerGroup") parameter. 
	     * @param customerGroup CustomerGroup for which the filter representation is to be added to the query's WHERE clause.
	     * @return This {@link Query} 
	     * @throws Exception
	     */
	    Query where(CustomerGroup customerGroup) throws Exception;

	    /**
	     * Parses and ANDs one or more SQL-like conditions to the query's WHERE clause.
	     * @param whereClauses SQL compatible conditions.
	     * @return This {@link Query} 	     */
	    Query where(String... whereClauses);
	    
	    /**
	     * Adds generic {@link Filter}s to the query's WHERE clause.
	     * @param filters One or more {@link Filter}s to be added to the query's WHERE clause.
	     * @return This {@link Query}
	     */
	    Query where(Filter... filters);

	    /**
	     * Parses and ANDs one or more SQL-like conditions to the query's HAVING clause.
	     * @param havingClauses SQL compatible conditions.
	     * @return This {@link Query} 	     */
	    Query having(String... havingClauses);

		/**
		 * Adds generic {@link Filter}s to the query's HAVING clause.
		 * <p>Example: 
		 * <pre>
		 * def threshold = api.getElement("ShowCountThreshold") ?: 0
		 *
		 * def dmCtx = api.getDatamartContext()
		 * def dmQuery = dmCtx.newQuery(dmCtx.getDatamart("SalesDM"))
		 *
		 * dmQuery.select("customerId")
		 * dmQuery.select("COUNT(Material)", "count")
		 * dmQuery.having(Filter.greaterThan("count", threshold))
		 *
		 * return dmCtx.executeQuery(dmQuery)?.getData()?.collect()
		 * </pre>
		 * Note: When aliases are defined in the dmQuery.select() calls, you need to use 
		 * these aliases in the dmQuery.having() calls.
		 * 
		 * @param filters One or more {@link Filter}s to be added to the query's HAVING clause.
		 * @return This {@link Query}
		 */
		Query having(Filter... filters);

		/**
		 * Parses and adds one or more SQL-like conditions to the query's ORDER BY clause.
		 * <p>Example:
		 * <pre>
		 * def dmCtx = api.getDatamartContext()
		 * def dmQuery = dmCtx.newQuery(dmCtx.getDatamart("SalesDM"))
		 *
		 * dmQuery.select("customerId")
		 * dmQuery.select("COUNT(Material)", "count")
		 * dmQuery.orderBy("count DESC")
		 *
		 * return dmCtx.executeQuery(dmQuery)?.getData()?.collect()
		 * </pre>
		 * Note: When aliases are defined in the dmQuery.select() calls, you need to use
		 * these aliases in the dmQuery.orderBy() calls.
		 * 
		 * @param orderClauses SQL compatible conditions. Sorting direction can be specified with a ' ASC' (default) or ' DESC' postfix.
		 * @return This {@link Query}
		 */
		Query orderBy(String... orderClauses);

	    /**
	     * Limits the number of rows returned in the query result. The absolute maximum is determined by the 'datamart.query.externalRowsLimit'
	     * Pricefx instance parameter (typically set to 100k rows).   
	     * @param maxRows Maximum number of rows to return. If not set, or set to 0, then the <code>externalRowsLimit</code> limit applies.
	     * @return This {@link Query}  
	     */
	    Query setMaxRows(Integer maxRows);

        /**
         * Allows for a number of query tweaking options, such as setting the target currency to convert to (if applicable) and requesting 
         * additional statistics to be calculated.
         * <p>Note: every call to this method fully replaces the previous options map, do not expect any merging.</p>
         *
         * @param options map of advanced query options
         * @return This {@link Query}
         */
        Query setOptions(Map<String, Object> options);

        /**
         * Allows to disable node level PA query caching just for this query.
         * <p>If query caching is disabled globally, this option cannot override that setting. So setting useCache to
         * true actually has no effect.</p>
         * @param useCache set to false if caching should be disabled for this query
         * @return This {@link Query}
         */
        Query setUseCache(Boolean useCache);

    }

	
	/**
	 * A DataSlice defines a set of filters to be applied to a PA DataSet (Datamart, DataSource...).
	 * A typical case is set of FC dimension (field,value) pairs.  In RebateManager a combination of a TimePeriod, CustomerGroup
	 * and ProductGroup is quite common. Complete generic {@link Filter}s can be used as well.
	 * Inclusions and exclusions are all OR-ed together. They are then combined as (inclusions AND NOT exclusions), i.e.
	 * <code>( (inclusion1 OR inclusion2 OR inclusion3...) AND NOT (exclusion1 OR exclusion2...) )</code>.
	 * <p>
	 * When a TimePeriod is used, the slice also needs to know the name of the date dimension field to match to.
	 * This field is set in the constructor and is common to all TimePeriods used.   
	 */
	interface DataSlice {
		
		void include(CustomerGroup customerGroup);

		void include(ProductGroup productGroup);
		
		void exclude(CustomerGroup customerGroup);

		void exclude(ProductGroup productGroup);

		void include(TimePeriod timePeriod);

		void exclude(TimePeriod timePeriod);
		
		void addFilter(Filter filter);
		
		void clear();
	}
		
	
    Object sourceSelectionEntry(String entryName, String... typeCode);
        
	/**
	 * DimFilter input parameter: renders a selection of all possible values for the given dimension field, in the
	 * FC which the column's table represents.
	 * 
	 * @param entryName Input param name.
	 * @param column Columns from the table representing the FC to get a dimension field value from.
	 * @return Selected dim field value.
	 */
	Object dimFilterEntry(String entryName, Column column);
	
	/**
	 * DimFilter input parameter: renders a selection of all possible values for the given dimension field, in the
	 * FC which the column's table represents.
	 * 
	 * @param entryName Input param name.
	 * @param column Columns from the table representing the FC to get a dimension field value from.
	 * @param defaultValue Value to use if no value has been selected yet.
	 * @return Selected dim field value or the default value if not yet set.
	 */
	Object dimFilterEntry(String entryName, Column column, String defaultValue);
	
	/**
	 * FC field selector, optionally limited to fields of a given type.
	 * 
	 * @param entryName Input param name.
	 * @param table Table that represents the FC to select a field from.
	 * @param sType Type of the field to allow the user to select:
	 * <p><ul>  
	 * 					<li> NUMBER </li> 
	 *					<li> QUANTITY  </li>
	 *					<li> TEXT  </li>
	 *					<li> DATE  </li>
	 *					<li> MONEY  </li>
	 *					<li> CURRENCY </li>
	 *					<li> UOM </li>
	 *					<li> LOB </li>
	 *					<li> DATETIME </li>
	 * </ul></p>
	 * @return Selected FC field.
	 */
	Object fieldSelectionEntry(String entryName, Object table, String sType);

	/** 
	 * FC field selector, optionally limited to fields of a given type.
	 * 
	 * @param entryName Input param name.
	 * @param table Table that represents the FC to select a field from.
	 * @param multiple Allow one field or multiple fields selection.
	 * @return Selected FC fields' names.
	 */
	Object fieldSelectionEntry(String entryName, Object table, String sType, Boolean multiple);
	
    Object fieldSelectionEntry(String entryName, String sourceName);

    Object fieldSelectionEntry(String entryName, String sourceName, Collection<String> sTypes, Boolean multiple);


    /**
     * Instantiates a new {@link DataLoader} to load rows of data into a {@link DMDataFeed} or {@link DMTable}.
     *
     * A {@link DataLoader} works in any logic context, as opposed to {@link DatamartRowSet} which is available only
     * in a PA DataLoad context.
     *
     * <p>Example:</p>
     * <pre>
     *     def ctx = api.getDatamartContext()
     *     def df = ctx.getDataFeed("Contract_Log")
     *     def loader = ctx.newDataLoader(df)
     *     loader.addRow(["ApprovalDate": date, "ApprovalUser": user])
     * </pre>
     *
     * @param table Table representing a {@link DMDataFeed} or {@link DMTable} to load data in.
     * @return {@link DataLoader} instance providing an API to add (buffer) and flush (commit to the DB) rows to a feed/table.
     */
    DataLoader newDataLoader(Object table);

    /**
     * Instantiates a new {@link DataLoader} to load rows of data into a {@link DMDataFeed} or {@link DMTable}.
     *
     * A {@link DataLoader} works in any logic context, as opposed to {@link DatamartRowSet} which is available only
     * in a PA DataLoad context.
     *
     * <p>Example:</p>
     * <pre>
     *     def ctx = api.getDatamartContext()
     *     def df = ctx.getDataFeed("Contract_Log")
     *     def loader = ctx.newDataLoader(df, "ApprovalDate", "ApprovalUser",...)
     *     loader.addRow(approvalDate, approvalUser, ...)
     * </pre>
     *
     * @param table Table representing a {@link DMDataFeed} or {@link DMTable} to load data in.
     * @param headerFieldNames Fields for which values will be loaded. Defaults to all (persisted) fields in the feed/table if not set.
     * @return {@link DataLoader} instance providing an API to add (buffer) and flush (commit to the DB) rows to a feed/table.
     */
    DataLoader newDataLoader(Object table, List<String> headerFieldNames);
	
	
	
    /**
     * API for loading data into a {@link DMDataFeed} or {@link DMTable} from a formula context.
     *
     * Safe to use outside of a PA DataLoad context.
     *
     * <p>Example:</p>
     * <pre>
     *     def ctx = api.getDatamartContext()
     *     def df = ctx.getDataFeed("Contract_Log")
     *     def loader = ctx.newDataLoader(df, "ApprovalDate", "ApprovalUser",...)
     *     loader.addRow(approvalDate, approvalUser, ...)
     * </pre>
     */
	interface DataLoader {
		
		/**
		 * Adds a row to the data buffer that will ultimately be loaded in the feed table.
		 * @param rowAsMap Row to add in a map of (field,value) pairs format.
		 */
		void addRow(Map<String,Object> rowAsMap);

		/**
		 * Adds a row to the data buffer that will ultimately be loaded in the feed table.		 
		 * @param rowAsList Row to add as a list of values matching the header fields order.
		 */
		void addRow(List<Object> rowAsList);

		/**
		 * Adds a row to the data buffer that will ultimately be loaded in the feed table.
		 * @param values Row to add as an array of values matching the header fields order.
		 */
		void addRow(Object... values);

		/**
		 * Flushes the data to the DB. This is in addition to the implicit flush that occurs when the internal buffer fills up.
		 * The internal buffer size is determined by the datamart.dataLoad.maxRowsPerBatch Pricefx node instance parameter, with 
		 * a default value of 20k rows. 
		 */
		void flush();
		
	}

	
	interface DataProfilerResult {

	    /**
	     * @return ResultMatrix with a row for each dimension profiled. Columns are:
	     * <p><ul>
	     *         <li> "Min" - Lowest value according to the natural order of the dimension field type (String, Date...). </li> 
	     *         <li> "Max" - Lowest value according to the natural order of the dimension field type (String, Date...). </li>
	     *         <li> "#" - Number of values in total (matching the number of rows in the query result set). </li>
	     *         <li> "#Nulls" - Number of rows with a null value (i.e. missing value). </li>
	     *         <li> "#Distinct" - Number of distinct values. </li>
	     *         <li> "Sample" - The first 10 such values. </li>
	     * </ul></p>        
	     */
        ResultMatrix getDimensions();

        /**
         * @return ResultMatrix with a row for each numeric projection profiled. Columns are:
         * <p><ul>
         *         <li> "Min" - Lowest value. </li>
         *         <li> "Max" - Highest value. </li>
         *         <li> "Mean" - Population's mean or average value. </li> 
         *         <li> "Std" - Population's standard deviation. </li>
         *         <li> "#" - Number of values in total (matching the number of rows in the query result set). </li>
         *         <li> "#Nulls" - Number of rows with a null value (i.e. missing value). </li>
         * </ul></p>        
         */
        ResultMatrix getMeasures();
        
        /** 
         * All the results in a map format, for fluent access in Groovy. 
         * @return  Map containing entries &lt;dim-name&gt;."Sample", &lt;measure-name&gt;."Mean" etc.
         */
        Map<String,Object> getMap();
	}
	
}
