package net.pricefx.formulaengine.scripting;

import java.io.Serializable;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import net.pricefx.common.api.pa.DataType;
import net.pricefx.server.dto.calculation.ResultMatrix;
import net.pricefx.service.DMUtil.FieldValues;

/**
 * Provides access to a QueryResult in a formula context, in the form of a two-dimension matrix.  The matrix columns correspond to query projections, matrix rows
 * correspond to the rows in the query result set, in the same order.  Row and column indexes are zero-based.
 * 
 * A query result may be obtained by
 * <p><ul> 
 * 		<li> executing a {@link net.pricefx.formulaengine.DatamartContext.Query Query} in the {@link net.pricefx.formulaengine.DatamartContext DatamartContext} </li>
 *      <li> executing a <code>org.apache.metamodel.query Query</code> in the {@link net.pricefx.formulaengine.TableContext TableContext} </li>
 * 		<li> calling the DatamartQuery function, or equivalent SandboxAPI api.datamartQuery() - both deprecated! </li>
 * </ul></p>
 */
public interface Matrix2D extends Serializable, Iterable<FieldValues> {

    /**
     * @return Number of rows in the matrix
     */
    public long getRowCount();

    /**
     * @return Number of columns in the matrix
     */
    public long getColumnCount();
    
    /**
     * @param col   Index of the column to get the label for.  The label is the corresponding projection's alias.
     * @return      The column label
     */
    public String getColumnLabel(long col);
    
    /**
     * A matrix is empty if it has no rows.
     * @return 
     */
    public boolean isEmpty();
    
    /**
     * @return  The column labels from left to right.
     */
    public List<String> getColumnLabels();
    
    /**
     * @param label Label of a column in the matrix
     * @return  The index of that column
     */
    public long getColumn(String label);

    /**
     * @param labels Labels of a set of columns in the matrix
     * @return  The indexes of those columns, in the same order
     */
    public long[] getColumns(Collection<String> labels);

    /**
     * @param colLabel  Label of the column to get the DataType for
     * @return  DataType of the given column; null if matrix does not have any meta defined.
     */
    public DataType getColumnDataType(String colLabel);

    /**
     * Selects one row in the matrix, and returns the result as new Matrix2D object, but which crucially does not duplicate any of the original rows.
     * Another way to look at this, is to see the result sub-matrix as a view on the original one.  Any changes to the row data in one matrix, will be reflected
     * in the other.
     * @param row   Index of row to select
     * @return      Sub-matrix containing that one row
     */
    public Matrix2D selectRow(long row);

    /**
     * Selects a subset of rows in the matrix, and returns the result as new Matrix2D object, but which crucially does not duplicate any of the original rows.
     * Another way to look at this, is to see the result sub-matrix as a view on the original one.  Any changes to the row data in one matrix, will be reflected
     * in the other.
     * @param startRow   Index of first row to select, inclusive
     * @param endRow     Index of last row to select, inclusive
     * @return           Sub-matrix containing all rows from the start to end row 
     */
    public Matrix2D selectRows(long startRow, long endRow);

    /**
     * Selects a column in the matrix, and returns the result as new Matrix2D object, but which crucially does not duplicate any of the original rows.
     * Another way to look at this, is to see the result sub-matrix as a view on the original one.  Any changes to the row data in one matrix, will be reflected
     * in the other.
     * @param label    Label of the column to select
     * @return         Sub-matrix containing that one column 
     */
    public Matrix2D selectColumn(String label);
    
    /**
     * Selects a column in the matrix, and returns the result as new Matrix2D object, but which crucially does not duplicate any of the original rows.
     * Another way to look at this, is to see the result sub-matrix as a view on the original one.  Any changes to the row data in one matrix, will be reflected
     * in the other.
     * @param col Index of the column to select
     * @return         Sub-matrix containing that one column 
     */
    public Matrix2D selectColumn(long col);

    /**
     * Selects a subset of columns in the matrix, and returns the result as new Matrix2D object, but which crucially does not duplicate any of the original rows.
     * Another way to look at this, is to see the result sub-matrix as a view on the original one.  Any changes to the row data in one matrix, will be reflected
     * in the other.
     * @param startCol   Index of first column to select, inclusive
     * @param endCol     Index of last column to select, inclusive
     * @return           Sub-matrix containing all columns from the start to end column 
     */
    public Matrix2D selectColumns(long startCol, long endCol);

    /**
     * Selects a subset of rows and columns in the matrix, and returns the result as new Matrix2D object, but which crucially does not duplicate any of the original rows.
     * Another way to look at this, is to see the result sub-matrix as a view on the original one.  Any changes to the row data in one matrix, will be reflected
     * in the other.
     * @param startRow   Index of first row to select, inclusive
     * @param endRow     Index of last row to select, inclusive
     * @param startColumn   Index of first column to select, inclusive
     * @param endColumn     Index of last column to select, inclusive
     * @return           Sub-matrix containing all cells from the rectangular area defined by the coordinates (startRow,startCol) to (endRow,endCol) 
     */
    public Matrix2D select(long startRow, long endRow, long startColumn, long endColumn);

    /**
     * Selects a subset of rows in the matrix, and returns the result as new Matrix2D object, but which crucially does not duplicate any of the original rows.
     * Another way to look at this, is to see the result sub-matrix as a view on the original one.  Any changes to the row data in one matrix, will be reflected
     * in the other.
     * @param rows      Array of row indices to select.  
     * @return          Sub-matrix containing the rows matching those indices, preserving the order from the original matrix 
     */
    public Matrix2D selectRows(long[] rows);

    /**
     * Selects a subset of rows in the matrix, and returns the result as new Matrix2D object, but which crucially does not duplicate any of the original rows.
     * Another way to look at this, is to see the result sub-matrix as a view on the original one.  Any changes to the row data in one matrix, will be reflected
     * in the other.
     * @param rows      Collection of row indices to select.  
     * @return          Sub-matrix containing the rows matching those indices, preserving the order from the original matrix 
     */
    public Matrix2D selectRows(Collection<? extends Number> rows);

    /**
     * Finds the index of row containing a given value in a given column, assuming this column is sorted high --&gt; low.
     * if there is no exact match we pick the previous or next one, whichever is closest to the given threshold value/
     * @param value     Threshold value 
     * @param col       Index of the column to search
     * @return          Index of the row closest to the threshold value
     */
    public long search(BigDecimal value, long col);

    /**
     * Find all rows for which the value in the given column is contained in the passed in collection.
     * This would typically be done only on columns containing discrete data.   
     * @param values    Collection of values to match to  
     * @param col       Index of the column to search
     * @return          Array of indices of matching rows 
     */
    public long[] search(final Collection<?> values, long col);
   
    /**
     * @return A new one row matrix, containing the maximum (of the population of values) for each numeric column in this matrix, at the same column indices. 
     * Non-numeric columns result in NaN. 
     */
    public Matrix2D max();

    /**
     * @return A new one row matrix, containing the minimum (of the population of values) for of each numeric column in this matrix, at the same column indices. 
     * Non-numeric columns result in NaN. 
     */
    public Matrix2D min();

    /**
     * @return A new one row matrix, containing the sum of the values for each numeric column in this matrix, at the same column indices. 
     * Non-numeric columns result in NaN. 
     */
    public Matrix2D sum();

    /**
     * @return A new one row matrix, containing the mean (of the population of values) for each numeric column in this matrix, at the same column indices. 
     * Non-numeric columns result in NaN. 
     */
    public Matrix2D mean();

    /**
     * @return A new one row matrix, containing the standard deviation (of the population of values) for each numeric column in this matrix, at the same column indices. 
     * Non-numeric columns result in NaN. 
     */
    public Matrix2D std();
    
    /**
     * @return A new one row matrix, containing the index of the row that holds the minimum value, for each numeric column in this matrix, at the same column indices. 
     * Non-numeric columns result in NaN. 
     */
    public Matrix2D indexOfMin();
    
    /**
     * @return A new one row matrix, containing the index of the row that holds the maximum value, for each numeric column in this matrix, at the same column indices. 
     * Non-numeric columns result in NaN. 
     */
    public Matrix2D indexOfMax();

    /**
     * Normalizes numeric column values, by dividing each value in the column by the total for the column.  If the total = 0, the normalized value is also 0.
     * @param cols  The indices of the columns to normalize.  These columns should contain numerical data, for th enormalization to be useful.
     * @return      A reference to this Matrix2D, in which the values have been normalized 'in situ'.
     */
    public Matrix2D normalize(long... cols);

    /**
     * Convenience getter of values in the matrix. 
     * @return  
     * <p><ul>
     *      <li> a single object value if the matrix has just one cell </li>
     *      <li> a map if the matrix as just one row </li>
     *      <li> a list of object values if the matrix has just one column </li>
     *      <li> a list of maps if the matrix has multiple rows and columns </li>
     *      <li> null for a 0-dim matrix (with 0 columns and or 0 rows) </li>
     *      <li> else just the matrix itself </li>
     * </ul></p>
     */
    public Object getValue();

    /**
     * @param row   Row index
     * @param col   Column index
     * @return      Value of cell at coordinates (row,col)
     */
    public Object getValue(long row, long col);

    /**
     * @param row   Row index
     * @return      Values on that row, in FieldValues (Map) format, i.e. [colLabel = value,...]
     */
    public FieldValues getRowValues(long row);

    /**
     * @param col   Column index
     * @return      The values in that column in List format, in the same order.
     */
    public List<Object> getColumnValues(long col);
    
    /**
     * Modifies the value at the given coordinates (row,col) 
     * @param val   Value to set
     * @param row   Row index  
     * @param col   Column index
     */
    public void setValue(Object val, long row, long col);

    /**
     * Modifies the value at the row coordinate, setting those values that match column names in the matrix.
     * @param val   Map of (column name, value) pairs.
     * @param row   Row index
     */
    public void setRowValues(Map<String,Object> val, long row);
    
    /**
     * Calculates the value at the given percentile in the given numeric column.
     * @param col           Column index
     * @param percentile    Integer number from 0 to 100
     * @return              The value so that percentile% of values (samples) in the column are smaller or equal to this value
     */
    public BigDecimal getPercentileValue(long col, int percentile);
    
    /**
     * Determines the total number of values (samples) smaller or equal to the given value, in the given numeric column.
     * @param col           Column index
     * @param value         Input to the CDF ('cumulative distribution function') 
     * @return              CDF evaluated at that value
     */
    public Long getCumulativeDistributionValue(long col, BigDecimal value);

    /**
     * Computes percentiles compatible with Excel
     * @param c             Column index
     * @return              0 to 100th percentile values in a 101-sized BigDecimal array (i.e. result[75] fives the 75th percentile).
     */
    public BigDecimal[] getExcelPercentileValues(long c);

    /**
     * Computes the rank of a value, as a percentage, in the distribution of values in the given column.
     * The values in the column do not need to be sorted upfront. The algorithm uses interpolation when there is no exact match
     * with the passed in value.
     * @param c  			Column index
     * @param value			The value for which to return the rank
     * @return				The rank, from 0 to 1, denoting the percentile of the distribution in which the given value falls. 
     */
    public double getExcelPercentRank(long c, double value);

    /**
     * Normalizes the values on a row-by-row basis.  The value for each included value is divided by the sum of those values.
     * If this sum is zero, the normalized values will also be zero.
     * @param colLabels     Labels of the columns to include in the normalization
     * @param precision     Maximum number of decimals in the normalized value 
     * @return              A reference to this Matrix2D, in which the values have been normalized 'in situ'.
     */
    public Matrix2D normalizeRowValues(Collection<String> colLabels, int precision);
   
    /**
     * @param label Label of a numeric column  
     * @return      Kurtosis of the distribution of values in that column
     */
    public double kurtosis( String label);
    
    /**
     * @param col   Column index
     * @return      A reference to this Matrix2D, in which the rows are now ordered so that the values in the given column are sorted from low --&gt; high
     */
    public Matrix2D sortRows(long col);

    /**
     * @param col   Column index
     * @return      A reference to this Matrix2D, in which the rows are now ordered so that the values in the given column are sorted from high --&gt; low
     */
    public Matrix2D sortRowsReverse(long col);

    /**
     * Extracts a map from the values in the nominated key and value columns. The assumption is the value column is functionally dependent on the key column
     * (as else key--&gt;value pairs on one row may overwrite the relationship between the pair on a previous row).
     *
     * @param keyColLabel   Label of the column for which the values are to be used as keys in the map
     * @param valColLabel   Label of the column for which the values are to be used as values in the map
     * @return              Key--&gt;value map in the same order as in this Matrix2D
     */
    public Map<String,Object> getMap(String keyColLabel,  String valColLabel);

    /**
     * Extracts a map from the values in the nominated key columns (resulting in keys of type MultiKey) and value columns.
     * The assumption is the value column is functionally dependent on the key column
     * (as else key--&gt;value pairs on one row may overwrite the relationship between the pair on a prevoius row).
     *
     * @param colLabels Labels of the columns for which the values are to be used as MultiKeys in the map
     * @param valColLabel Label of the column for which the values are to be used as values in the map
     * @return MutltiKey--&gt;value map in the same order as in this Matrix2D
     */
    public Map<Object, Object> getMap(final List<String> colLabels, final String valColLabel);
        
    /**
     * Add a column to this Matrix2D, with the value on each row looked up from the given map, using the mapped column's value as the lookup key. 
     * @param colLabel          Label of the new column
     * @param mappedColLabel    Label of the column holding the lookup keys    
     * @param mapping           Map holding the values to map those lookup keys to
     * @return                  A reference to this matrix, to which the new column is added
     */
    public Matrix2D addMappedColumn(String colLabel, String mappedColLabel, Map<String,Object> mapping);
    
    /**
     * Add a column to this Matrix2D, with the value on each row looked up from the given map, using a set of existing columns' values as a composite lookup key.
     * The keys in the map are assumed to have been instantiated by {@link PublicGroovyAPI#multiKey(Object...)}.
     * @param colLabel          Label of the new column
     * @param fromColLabels     Labels of the columns holding the composite lookup key values    
     * @param mapping           Map holding the values to map those lookup keys to
     * @return                  A reference to this matrix, to which the new column is added
     */
    public Matrix2D addMappedColumn(String colLabel, List<String> fromColLabels, Map<String,Object> mapping);
    
    /**
     * Add a column to this Matrix2D, with the value on each row retreived from the given values array.
     * @param colLabel          Label of the new column
     * @param colValues         Array of values to populate the new column with
     * @return                  A reference to this matrix, to which the new column is added
     */
    public Matrix2D addColumn( String colLabel,  Object[] colValues);

    /**
     * Add a column to this Matrix2D, without any values
     * @param colLabel          Label of the new column
     * @return                  A reference to this matrix, to which the new column is added
     */
    public Matrix2D addColumn( String colLabel);

    /**
     * Renames the columns in this matrix so that the old column label = key, and the new column label = value in the given map.
     * @param old2NewColLabels  Old --&gt; new column label mapping
     * @return                  A reference to this matrix, in which the matching columns have been renamed
     */
    public Matrix2D renameColumns(Map<String,String> old2NewColLabels);
    
    /**
     * Replaces nulls in the given column with the given value.
     * @param colLabel  Label of column to process
     * @param val       Value to set
     * @return          A reference to this matrix, in which the null values have been replaced.
     */
    public Matrix2D replaceMissingValues(String colLabel, Object val);

    public enum BucketNaming { ABC, INT, RANGE, RANGE_PCT };

    /**
     * Convenience method to quantize all numeric columns in one go. 
     * @param numberOfBuckets
     * @param newColLabelPostfix Example: the bins for column 'margin' go into a new column named 'margin_q' if the prefix is set to '_q'.  
     * @param bucketNaming
     */
    public List<String> quantize(int numberOfBuckets, String newColLabelPostfix, String bucketNaming);

    /**
     * Performs equal width quantization of a scalar distribution in a given column, and maps each value to its nearest lower quantile.
     * A column is added if none with the given label exists, or when the label is null.
     * @param colLabel          Label of the column holding the scalar values
     * @param numberOfBuckets   Number of quantiles/buckets to calculate (max 100).  Typically 10 deciles or 4 quartiles.
     * @param newColLabel       Label of the new column to add, holding the mapped quantile
     * @param bucketNaming      One of the following values:
     * <p><ul>
     *                          <li> "ABC":        quantiles are named "A","B","C" etc. </li>
     *                          <li> "INT":        quantiles are numbered 0,2,3,... </li>
     *                          <li> "RANGE":      quantile names reflect the range of the bucket (lower bound → upper bound) </li>
     *                          <li> "RANGE_PCT":  as "RANGE" but not the bounds are formatted as percentages (ex. 8% vs 0.08) </li>
     * </ul></p>                         
     */
    public void quantize(String colLabel, int numberOfBuckets, String newColLabel, String bucketNaming);
    
    public ResultMatrix toResultMatrix();

    /**
     * Converts this Matrix2D into ResultMatrix
     *
     * @param withColumnFormats boolean flag to add columns formats and normalize datetime fields.
     * @return ResultMatrix
     * @see ResultMatrix#normalizeDateTimes()
     */
    public ResultMatrix toResultMatrix(boolean withColumnFormats);

    
    /**
     * Create a ResultMatrix from a Matrix2D, selecting only the columns included in the mapping if such a mapping is provided.
     * 
     * @param columnMapping Mapping ResultMatrix column name --&gt; Matrix2D column label
     * @return ResultMatrix
     */
    public ResultMatrix toResultMatrix(Map<String,String> columnMapping);

    public ResultMatrix toResultMatrix(Map<String,String> columnMapping, boolean withColumnFormats);
    
    
	public static Matrix2D fromResultMatrix(ResultMatrix resultMatrix) {
        return null;
	}

    public static Matrix2D fromListOfMaps(final Collection<Map<String,Object>> rows) {
        return null;
    }

    public static Matrix2D fromListOfMaps(final Collection<String> colLabels, final Collection<Map<String,Object>> rows) {
        return null;
	}

}