package net.pricefx.lmformulaengine.scripting;

import com.fasterxml.jackson.core.type.TypeReference;
import com.googlecode.genericdao.search.Filter;
import groovy.lang.Closure;

import java.io.IOException;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;

import net.pricefx.common.api.InputType;
import net.pricefx.domain.PriceGrid;
import net.pricefx.domain.Pricelist;
import net.pricefx.formulaengine.AbstractProducer.ResultIterator;
import net.pricefx.formulaengine.DatamartContext;
import net.pricefx.formulaengine.DatamartRowSet;
import net.pricefx.formulaengine.TableContext;
import net.pricefx.formulaengine.XExpression;
import net.pricefx.formulaengine.scripting.inputbuilder.InputBuilderFactory;
import net.pricefx.lmformulaengine.LmDebugUnsupportedException;
import net.pricefx.lmformulaengine.LmFormulaEngineContext;
import net.pricefx.lmformulaengine.Util;
import net.pricefx.lmformulaengine.model.MapExtended;
import net.pricefx.logic.core.TypeCode;
import net.pricefx.logic.core.Utils;
import net.pricefx.logic.core.model.FormulaTraceMessage;
import net.pricefx.pckg.transform.descriptor.TypeDescriptor;
import net.pricefx.restapiclient.PfxClient;
import net.pricefx.restapiclient.PfxClientException;
import net.pricefx.server.dto.calculation.*;
import net.pricefx.service.ItemSummaryQuery;
import net.pricefx.service.RebateRecordSummaryQuery;
import net.pricefx.service.SimulationItemSummaryQuery;
import net.pricefx.service.impl.AdvancedFilterUtil;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.lang3.ArrayUtils;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static net.pricefx.common.api.PriceFxInterface.*;


/**
 * This interface represents what can be reached in groovy formula syntax as
 * {@code api}. That {@code api} variable or binding is merely an instance of
 * this interface to expose certain "safe" methods and functionality.<br />
 * <br />
 * <b>Caching usage</b><br />
 * The api exposes 3 ways to store transient and temporary data during formula
 * executions. Each variant has its own specifics and therefore is not suitable for all use cases. The options are:<br />
 * <ul>
 * <li>api.local</li>
 * <li>api.global</li>
 * <li>Shared Cache</li>
 * </ul>
 * <br />
 * The first two can be understood as mere in-memory hash maps. For example,
 * api.local.put("key","value") will work fine.<br />
 * The difference between them is the potential scope and retention of the
 * contained data.<br />
 * <br />
 * <u>api.local:</u><br />
 * This is a hash map that is available during the lifetime of exactly one
 * formula execution. It can (and should) be used to store intermediate results
 * that need to be accessed by other subsequent formula elements.<br />
 * <br />
 * <u>api.global:</u><br />
 * By default, api.global is very similar to api.local. Without any further
 * settings, the only difference is that api.global is also available in Groovy
 * library functions (while api.local is not).<br />
 * However, by setting api.retainGlobal = true you can instruct the formula
 * engine to keep the hash map in between formula runs in case e.g. a price list
 * is calculated. The values would be available for the next line item's
 * calculation right away. The values are also carried forward between two list
 * calculation passes (i.e. form initial run to dirty item run; ONLY in non-distributed mode!). <b><br />
 * Important notes:
 * <br />retainGlobal is effective only once. In order to keep it for
 * the 3rd item as well (and so on) you need to call it in every formula
 * execution.<br />
 * api.global data also sticks to its execution thread. That means that it is NOT shared between
 * the different parallel executions of a distributed calculation. It effectively builds up multiple
 * caches. Another side effect of this is that api.global is not retained and passed to a subsequent 
 * calculation pass if the original calculation was in a distributed mode.
 * <br /></b>
 * <br />
 * <u>Shared Cache:</u><br />
 * The main limitation of api.global is the restriction of its availability to
 * the current execution thread. In a distributed calculation however, there are
 * multiple threads and even multiple servers involved in the calculation. If all
 * these should share a single cache, you must use the shared cache. The shared
 * cache is even available to different calculations (such as other lists or list &
 * CFS or dashboard formula etc). Be careful with key naming so that you do not
 * mix and match data.<br />
 * These advantages come at a cost: Due to the distributed "superglobal" nature,
 * the read/write operations of that cache are slower compared to a real
 * local in-memory map. Use the shared cache only for data that is expensive
 * to calculate or for avoiding frequent database calls. Data in the shared cache
 * expires automatically, the TTL is extended everytime a key is written or read.
 * Values stored are "String" type only. Serialization and deserialization of other
 * objects like dates and numbers are to be done in formula itself. 
 * <br /><br />
 * <b>Object modifications</b><br />
 * The API does contain methods to manipulate data (add, addOrUpdate, delete, etc).
 * These methods can only be used (and will otherwise be ignored) in certain formula contexts.
 * These contexts are: <br/>
 * <ul>
 * <li>From within a CFS execution (non distributed only!)</li>
 * <li>From within a calculation flow</li>
 * <li>From within a direct formula execution via JSON API </li>
 * </ul>
 * <br />
 * On top of this, some restrictions apply which data objects can be manipulated - i.e. not every possible object can be modified.
 * Currently available whitelisted typecodes are:
 * <ul>
 * <li>All customer extension types</li>
 * <li>All product extension types</li>
 * <li>All matrix price parameter values and normal price parameter values (LTV) as well as price parameter header (LT)</li>
 * <li>PGI and XPGI</li>
 * <li>PLI and XPLI</li>
 * <li>MPLI</li>
 * <li>P</li>
 * <li>C</li>
 * <li>PCOMP</li>
 * <li>PDESC</li>
 * <li>PGI</li>
 * <li>TODO</li>
 * <li>Data export, import and archive</li>
 * </ul>
 */
public class LmSandboxAPI {

    private static final Logger log = LoggerFactory.getLogger(LmSandboxAPI.class);

    protected transient LmFormulaEngineContext ctx;

    public transient Map<String, Object> global = new HashMap();
    public transient Map<String, Object> local = new HashMap();
    private transient Map<String, String> sharedCache = new HashMap();

    public boolean retainGlobal;
    private LmSandboxApiAgent lmApi;
    private String partitionName;

    public LmSandboxAPI(PfxClient client, TypeDescriptor objectType, LmFormulaEngineContext ctx) {
        this.ctx = ctx;
        this.partitionName = client.getPartition();
        lmApi = new LmSandboxApiAgent(client, objectType, ctx.getParametersClone());
    }

    /**
     * Returns true if the formula is executed in syntax check (or parameter determination) mode.
     */
    boolean isSyntaxCheck() {
        return false;
    }

    /**
     * Returns true if formula is called via UI debug/test exec mode.
     *
     * @return true if executed in debug mode
     */
    boolean isDebugMode() {
        return true;
    }

    /**
     * Returns true if calculation is executed in distributed mode.
     * Useful to suppress warnings (or not execute methods) for operations that do not work in distributed mode.
     * These operations are basically all writing operations that affect other items.
     * Examples are: all object modification ops (add, addOrUpdate, update, delete) and marking other items dirty
     *
     * @return true if executed in distributed mode
     */
    boolean isDistributedMode() {
        return false;
    }

    /**
     * A publicly available cache (as a hashmap) for storing and sharing values across formula elements.
     * If <code>api.retainGlobal</code> is set to true, this map will also be available with the entire list calculation,
     * i.e across item boundaries. It can be also accessed directly as <code>api.global</code>.
     *
     * <strong>Important</strong>: This map must be treated as a cache.
     * It means that the formula should never rely on the cached values, as in various situations this cache is purged
     * or structured differently. For example, in a distributed calculation, every calculation thread has its own cache;
     * or in a multi-pass list calculation, the subsequent passes may not have the cache of the previous pass available.
     * The system always tries to make it available, but the formula should never rely on it.
     *
     * @return the hashmap
     */
    Map<String, Object> getGlobal() {
        return global;
    }

    /**
     * Puts an entire map of variables into the global cache.
     *
     * @param globalVars The map of variables to put into the global cache
     */
    void putGlobal(final Map<String, Object> globalVars) {
        global.clear();
        if (globalVars != null){
            global.putAll(globalVars);
        }
    }

    /**
     * Retreive the CalculationContext of the formula evaluation (aka formula element group).<br />
     * <ul>
     * <li>a) For a PA Calculation or Flush formula this can be (from PriceFxInterface):<ul>
     *      <li>- FORMULAELEMENTGROUP_DM_INIT = "init";</li>
     *      <li>- FORMULAELEMENTGROUP_DM_ROW = "row";</li>
     *      <li>- FORMULAELEMENTGROUP_DM_SUMMARY = "summary";</li></ul>
     * </li>
     * <li>b) In RebateAgreement calculation we have:<ul>
     *      <li>- FORMULAELEMENTGROUP_RM_AGREEMENT = "agreement";</li>
     *      <li>- FORMULAELEMENTGROUP_RM_AGREEMENT_READONLY = "agreementReadOnly";</li>
     *      <li>- FORMULAELEMENTGROUP_RM_REBATERECORD = "rebateRecord";</li></ul>
     * </li>
     * <li>c) In the Optimizer module, when evaluating Model formulas, there is:<ul>
     *      <li>- FORMULAELEMENTGROUP_MODEL_DEFINITION = "definition";</li>
     *      <li>- FORMULAELEMENTGROUP_MODEL_CALCULATION = "calculation";</li>
     *      <li>- FORMULAELEMENTGROUP_MODEL_ANALYSIS = "analysis";</li></ul>
     * </li></ul>
     * @return CalculationContext as String
     */
    String getCalculationContext() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Gets the info about the current batch of items being calculated by a calcuation thread.
     * The calculation batch size is defined by a server variable calculationTasks.commitBatchSize and by default it is 200 records.
     * For the best use, use the functions from the <a href="https://bitbucket.org/pricefx/pricefx-logic/src/master/CalculationLogic/SharedLib/elements/BatchUtils.groovy">Shared Library</a>.
     *
     * @return List of arrays [sku, key2] for PLI, XPLI, PGI, XPGI contexts or [itemId, typedId] for CFS logics run on P, PX
     * @see <a href="https://bitbucket.org/pricefx/pricefx-logic/src/master/CalculationLogic/SharedLib/elements/BatchUtils.groovy">SharedLib/elements/BatchUtils.groovy</a>
     */
    public List<String[]> getBatchInfo() {
        return ctx.getBatchInfo();
    }

    /**
     * Returns the id part of an object. The method tries to be smart and extracts also ids from maps.
     * or typedId strings
     */
    Long getId(Object obj) {
        if(obj == null) return null;

        if(obj instanceof Map) {
            if(((Map)obj).containsValue(FIELD_ID)) {
                return Long.valueOf(((Map)obj).get(FIELD_ID).toString());
            }
            if(((Map)obj).containsValue(FIELD_TYPEDID)) {
                return Utils.convertTypedIdToId(((Map) obj).get(FIELD_ID).toString());
            }
        }

        if(obj instanceof String) {
            return Utils.convertTypedIdToId((String) obj);
        }

        return null;
    }

    /**
     * Gets the result of a previous formula element.
     *
     * @param name Name of the formula element
     * @return The value or {@code null} if the element is not found
     */
    Object getElement(String name) {
        return this.ctx.getResultValue(name);
    }

    /**
     * Gets the secondary key in case a matrix pricelist or matrix simulation is processed.
     *
     * @return The value of the secondary key
     */
    Object getSecondaryKey() {
        return this.ctx.getSecondaryKey();
    }

    /**
     * Gets a named binding variable from a library function.
     * Example:
     * <p>
     * In a workflow formula the workflow can be referenced in a groovy element directly by the binding/variable "workflow":
     * workflow.addApprovalStep("First Approval")
     * <p>
     * However this does not work from a library function as this script does not have the binding available.
     * This would work:
     * api.getBinding("workflow").addApprovalStep("First Approval")
     */
    Object getBinding(String name) {
        return this.ctx != null && this.ctx.getGroovyBindings() != null?this.ctx.getGroovyBindings().get(name):null;
    }

    /**
     * Gets the current execution context's target date.
     *
     * @return The date
     */
    Date targetDate() {
        return ctx.getTargetDate();
    }

    /**
     * Gets a calendar object initialized with the current targetDate (or now if no target date is given).
     *
     * @return The initialized Calendar object
     */
    Calendar calendar() {
        return calendar(null);
    }

    /**
     * Gets a calendar object initialized with <code>date</code>.
     *
     * @param date The date to initialize the calendar with
     * @return The initialized Calendar object
     */
    Calendar calendar(Date date) {
        GregorianCalendar c = new GregorianCalendar();
        if(date == null) {
            if(this.ctx.getTargetDate() == null) {
                date = new Date();
            } else {
                date = this.ctx.getTargetDate();
            }
        }

        c.setTime(date);
        return c;
    }

    /**
     * @return The <code>java.util.Random</code> generator instance created for the duration of the calculation task.
     * (i.e.) does not get recreated for each formula (element) evaluation).
     */
    Random random() {
        return this.ctx.getRandom();
    }

    /**
     * @return A {@link net.pricefx.domain.DMCalendar DatamartCalendar} object.
     * @deprecated Use {@link net.pricefx.formulaengine.DatamartContext#calendar} instead.
     */
    @Deprecated
    Object datamartCalendar() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates a new DatamartSlice, which allows for setting filter criteria along the Time, CustomerGroup, ProductGroup
     * or any other dimensions in a Datamart.
     *
     * @return The instantiated, empty, {@link DataSlice} object.
     * @deprecated Now provided by {@link net.pricefx.formulaengine.DatamartContext DatamartContext}.
     */
    @Deprecated
    Object newDatamartSlice() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates a new DatamartSlice, which allows for setting filter criteria along the Time, CustomerGroup, ProductGroup
     * or any other dimensions in a Datamart, intiialized with the name of the time dimension field, and an optional set
     * filter criteria.
     *
     * @param dateFieldName Name of the time dimension field.
     * @param timePeriodsAndProductAndCustomerGroups TimePeriod, CustomerGroup, ProductGroup filters.
     * @return The initialzied {@link DataSlice} object.
     * @deprecated Now provided by {@link net.pricefx.formulaengine.DatamartContext DatamartContext}.
     */
    @Deprecated
    Object newDatamartSlice(String dateFieldName, Object... timePeriodsAndProductAndCustomerGroups) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Translates a CustomerGroup/ProductGroup/DMDataSlice to its equivalent Filter representation,
     * mapping domain level field names to the corresponding Datamart field names in the process.
     * <p>
     * Note: A DMDataSlice object can be instantiated by api.newDatamartSlice, or in a Rebate calculation context, referenced by the 'calulationBase' binding.
     *
     * @param groupOrSlice The CustomerGroup/ProductGroup/DMDataSlice to translate into Datamart level filter criteria
     * @return Filter object representing either the CustomerGroup/ProductGroup dimension slice, or the passed slice in the Datamart
     */
    Filter datamartFilter(Object groupOrSlice) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Important: please use the Groovy query API instead (see {@link net.pricefx.formulaengine.DatamartContext DatamartContext}).
     * This method only exists for backward compatibility for older in-production formulas.
     *
     * Runs a query against a PA datamart. The query is only executed once on the scope of a calculation job (PL generation etc.).
     * A query is identified by the first argument. This is also the 'handle' by which datamartLookup refers to the result set of the query.
     *
     * @param queryAndDatamartAndProjectionsAndFilters  First argument is the query ID/handle.  Subsequent arguments are either DM field names (projections)
     * or filter objects.
     * @return The numbers of rows in the result set.  The actual result set data is stored in memory for subsequent lookup by datamartLookup(...).
     */
    Object datamartQuery(Object... queryAndDatamartAndProjectionsAndFilters) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Important: Please use the Groovy query API instead (see {@link net.pricefx.formulaengine.DatamartContext DatamartContext}).
     * This method only exists for backward compatibility for older in-production formulas.
     *
     * Looks up rows in the result of a datamartQuery(...) call.
     * @param queryAndProjectionsAndFilters  The first arguments is the ID/handle from a datamartQuery(...) call.  Subsequent arguments refer to
     * projections specified in the datamartQuery call, or filter objects leading to additional filtering of rows in the result set.
     * @return A single primitive value if the result of the lookup results in one row for just one projection.
     * A list of values if the result is multiple rows for one projection.
     * A map of (field name, value) pairs for a single row result with multiple projections.
     * Finally, a Matrix2D object in cse of multiple rows and projecitons.
     */
    Object datamartLookup(Object... queryAndProjectionsAndFilters) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }


    /**
     * The DatamartContext provides an API for querying and loading PA data.
     * @return {@link net.pricefx.formulaengine.DatamartContext DatamartContext}
     */
    DatamartContext getDatamartContext() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * The TableContext provides an SQL query (SELECT only) interface to PA query results.
     * @return {@link net.pricefx.formulaengine.TableContext TableContext}
     */
    TableContext getTableContext() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Available in a PA DataLoad context only, a {@link DatamartRowSet} provides access to the rows being loaded or calculated.
     * @return The {@link DatamartRowSet} holding the data rows to load in the target FC.
     */
    DatamartRowSet getDatamartRowSet() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Available in a PA DataLoad context only, a {@link DatamartRowSet} provides access to the rows currently being loaded or calculated.
     * The following DL types make use of rowsets:<ul>
     * 		<li>- 'Flush': The source rowset gives access to the DataFeed data rows, while the target rowset holds the rows to be loaded in the target DataSource (DS).</li>
     * 		<li>- 'Calculate': The source rowset gives access to the source rows, which can either come from a DF or from the target FC being calculated.
     * 						The target rowset holds the rows to be loaded in the target FC. Initially, this rowset will be empty unless the DL.withTargetSnapshot
     * 						option is set, in which case the rowset will be pre-populated with the target FC rows in the scope of the DL (determined by the DL.filter).</li></ul>
     * @param name Either 'source' or 'target', to select the source FC or target FC rowset.
     * @return The {@link DatamartRowSet} holding either the source or target rowset.
     */
    DatamartRowSet getDatamartRowSet(String name) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Temporarily switches the current sku context to another sku.
     * This can be used to call functions that rely on a sku context and do not allow to specify an alternate sku with a different sku.
     * The switch is reset by {@link #resetSKUContextSwitch} or automatically.
     *
     * @param newSKU The new sku to switch to.
     */
    void switchSKUContext(String newSKU) {
        this.ctx.switchSKUContext(newSKU);
    }

    /**
     * Resets a sku context switch to the origionally specified sku.
     */
    void resetSKUContextSwitch() {
        this.ctx.resetSKUContextSwitch();
    }

    /**
     * Returns the current product object. This one does not exist in the regular syntax, but is useful in Groovy.
     * @return The full current product
     */
    Map<String, Object> product() throws PfxClientException {
        Map<String, Object> product = ctx.getProduct();
        if (product == null) {
            product = lmApi.callApi();
            ctx.setProduct(product);
        }
        return product;
    }

    public InputBuilderFactory inputBuilderFactory() {
        return new InputBuilderFactory(ctx, null);
    }

    /**
     * Equivalent of the standard syntax "Product" function. Retrieves the given attribute's value of the current product.
     *
     * @param attributeName The attribute whose value to return
     * @return The value of the attribute
     */
    Object product(String attributeName) throws PfxClientException {
        Map<String, Object> product = product();
        return product != null ? product.get(attributeName) : null;
    }

    /**
     * Equivalent of the standard syntax "Product" function. Retrieves the given attribute's value of the given product.
     *
     * @param attributeName The attribute whose value to return
     * @param sku           The product
     * @return The value of the attribute (or {@code null} if product not found)
     */
    Object product(String attributeName, String sku) throws PfxClientException {
        return lmApi.callApi(attributeName, sku);
    }

    /**
     * Retrieves the given attribute value from the product table for a given SKU.
     * If the given SKU is null, filterFormulaName and filterFormulaParam are used to apply
     * filter logic for the Product picker.
     * @param attributeName 		The attribute whose value should be returned
     * @param sku					The product
     * @param filterFormulaName  	Unique name of a formula that will be triggered when the Product picker opens
     * @param filterFormulaParam 	Additional data for filterFormula
     * @return The value of the attribute (or {@code null} if product not found)
     */
    Object product(String attributeName, String sku, String filterFormulaName, Object filterFormulaParam) throws PfxClientException {
        return lmApi.callApi(attributeName, sku, filterFormulaParam);
    }

    /**
     * A method to trigger a custom-named product picker input parameter.
     *
     * @param parameterName The name of the parameter
     * @param attributeName The attribute which value to return ({@code null} for the entire object)
     * @return The attribute value
     */
    Object otherProduct(String parameterName, String attributeName) throws PfxClientException {
        return lmApi.callApiUsingFormula(parameterName, attributeName);
    }

    /**
     * A method to trigger a custom-named product picker input parameter.
     *
     * @param parameterName The name of the parameter
     * @return The product object
     */
    Object otherProduct(String parameterName) throws PfxClientException {
        return otherProduct(parameterName, null);
    }

    /**
     * Equivalent of the standard syntax "PXLookup" function. Retrieves a list of PX entries for the current product.
     *
     * @param extensionName The name of the extension
     * @return A list of matching PX records
     */
    Object productExtension(String extensionName) throws PfxClientException {
        Object result = lmApi.callApi(extensionName);
        if (result instanceof List) {
            return MapExtended.convertList((List<Map<String, Object>>) result, TypeCode.PRODUCTEXTENSION.getShortName(), extensionName);
        }
        return result;
    }

    /**
     * Equivalent of the standard syntax "PXLookup" function. Retrieves a list of PX entries for the current product.
     *
     * @param extensionName The name of the extension
     * @param filters       One or more {@link #filter} objects that narrow down the PX search
     * @return A list of matching PX records
     */
    Object productExtension(String extensionName, Filter... filters) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{extensionName}, (Object[]) filters);
        Object result = lmApi.callApi(params);
        if (result instanceof List) {
            return MapExtended.convertList((List<Map<String, Object>>) result, TypeCode.PRODUCTEXTENSION.getShortName(), extensionName);
        }
        return result;
    }

    /**
     * Equivalent of the standard syntax "Competition" function. Retrieves a list of competition records for the current product.
     *
     * @return A list of PCOMP records
     */
    Object productCompetition() throws PfxClientException {
        //TODO: after backend call supports empty parameters
        return find(TypeCode.PRODUCTCOMPETITION.getShortName(), Filter.equal(FIELD_SKU, product(FIELD_SKU)));
    }

    /**
     * Equivalent of the standard syntax "Competition" function. Retrieves a list of competition records for the current product.
     *
     * @return * @param filters One or more {@link #filter(java.lang.String, java.lang.Object)} objects that narrow down the competition search
     */
    Object productCompetition(Filter... filters) throws PfxClientException {
        Object result = lmApi.callApi((Object[]) filters);
        if (result instanceof List) {
            return MapExtended.convertList((List<Map<String, Object>>) result, TypeCode.PRODUCTCOMPETITION.getShortName());
        }
        return result;
    }

    /**
     * Equivalent of the standard syntax "XRefLookup" function. Retrieves a list of x-reference records for the current product.
     *
     * @return A list of PXREF records
     */
    Object productXRef() throws PfxClientException {
        //TODO: after backend call supports empty parameters
        return find(TypeCode.PRODUCTXREFERENCE.getShortName(), Filter.equal(FIELD_SKU, product(FIELD_SKU)));
    }

    /**
     * Prompts user for rebate agreement.
     *
     * @param parameterName - (mandatory) name of entry in input parameters listing
     * @param attributeName - (mandatory) name of attribute you need
     * @return attribute of rebate agreement or null if RA does not exists
     */
    Object rebateAgreementUserEntry(String parameterName, String attributeName) {
        //TODO
        return lmApi.getParameter(parameterName);
    }

    /**
     * Equivalent of the standard syntax "XRefLookup" function. Retrieves a list of x-reference records for the current product.
     *
     * @param filters One or more {@link #filter(java.lang.String, java.lang.Object)} objects that narrow down the PXREF search
     */
    Object productXRef(Filter... filters) throws PfxClientException {
        Object result = lmApi.callApi((Object[]) filters);
        if (result instanceof List) {
            return MapExtended.convertList((List<Map<String, Object>>) result, TypeCode.PRODUCTXREFERENCE.getShortName());
        }
        return result;
    }

    /**
     * Equivalent of the standard syntax "Customer" function. Retrieves the given attribute's value of the customer context parameter.
     *
     * @param attributeName The attribute whose value to return
     * @return The value of the attribute
     */
    Object customer(String attributeName) throws PfxClientException {
        Map<String, Object> customer = ctx.getCustomer();
        if (customer == null) {
            String customerId = (String) ctx.getParameter(CONTEXTPARAM_NAME_CUSTOMER);
            if (customerId != null) {
                List<Map<String, Object>> customers = find(TypeCode.CUSTOMER.getShortName(), 0, 1, null, Filter.equal("customerId", customerId));
                if (!customers.isEmpty()) {
                    customer = customers.get(0);
                    ctx.setCustomer(customer);
                }
            }
        }

        return customer != null ? customer.get(attributeName) : null;
    }

    /**
     * Equivalent of the standard syntax "Customer" function. Retrieves the given attribute's value of the customer context parameter.
     * If the value is <code>null</code> and <em>defaultFromParent</em> is <code>true</code>, it retrieves (and returns) the value from
     * the customer's parent no matter if that value is <code>null</code> as well or not. In case there's no parent customer set,
     * <code>null</code> will be returned.
     *
     * @param attributeName     The attribute whose value to return
     * @param defaultFromParent Tells whether to lookup the value from the parent customer (if any) in case it's <code>null</code> in the customer
     * @return The value of the attribute
     */
    Object customer(String attributeName, boolean defaultFromParent) throws PfxClientException {
        return lmApi.callApiUsingFormula(attributeName, defaultFromParent);
    }

    /**
     * Equivalent of the standard syntax "Customer" function. Retrieves the given attribute's value of the given customer.
     *
     * @param attributeName The attribute whose value to return
     * @param customerId    The customerId value for the customer search
     * @return The value of the attribute (or {@code null} if no customer found)
     */
    Object customer(String attributeName, String customerId) throws PfxClientException {
        return lmApi.callApi(attributeName, customerId);
    }

    /**
     * Equivalent of the standard syntax "Customer" function. Retrieves the given attribute's value of the given customer. If the value is
     * <code>null</code> and <em>defaultFromParent</em> is <code>true</code>, it retrieves (and returns) the value from the customer's
     * parent no matter if that value is <code>null</code> as well or not. In case there's no parent customer set, <code>null</code> will be returned.
     *
     * @param attributeName     The attribute whose value to return
     * @param customerId        The customerId value for the customer search
     * @param defaultFromParent Tells whether to lookup the value from the parent customer (if any) in case it's <code>null</code> in the customer
     * @return The value of the attribute (or {@code null} if no customer found)
     */
    Object customer(String attributeName, String customerId, boolean defaultFromParent) throws PfxClientException {
        return lmApi.callApi(attributeName, customerId, defaultFromParent);
    }

    /**
     * Equivalent of the standard syntax' "Customer" function. Retrieves a named attribute's value of the named customer (if the value is
     * <code>null</code> and <em>defaultFromParent</em> is <code>true</code>, retrieves (and returns) the value from the customer's
     * parent no matter if that value is <code>null</code> as well or not. In case there's no parent customer set, <code>null</code> will be returned)
     *
     * @param attributeName      The attribute whose value to return
     * @param customerId         The customerId value for the customer search
     * @param defaultFromParent  Tells whether to lookup the value from the parent customer (if any) in case it's <code>null</code> in the customer
     * @param filterFormulaName  Unique name of a formula, that will be triggered when Customer Picker opens
     * @param filterFormulaParam Additional data for filterFormula (e.g. agreement's typedId,...)
     * @return The value of the attribute (or {@code null} if no customer found)
     */
    Object customer(String attributeName, String customerId, boolean defaultFromParent, String filterFormulaName, Object filterFormulaParam) throws PfxClientException {
		return lmApi.callApi(attributeName, customerId, defaultFromParent, filterFormulaName, filterFormulaParam);
	}

    /**
     * A method to trigger a custom-named customer picker input parameter.
     *
     * @param parameterName The name of the parameter
     * @return The customer object
     */
    Object otherCustomer(String parameterName) throws PfxClientException {
        return otherCustomer(parameterName, null);
    }

    /**
     * A method to trigger a custom-named customer picker input parameter.
     *
     * @param parameterName The name of the parameter
     * @param attributeName The attribute which value to return ({@code null} for the entire object)
     * @return The attribute value
     */
    Object otherCustomer(String parameterName, String attributeName) throws PfxClientException {
        return otherCustomer(parameterName, attributeName, false);
    }

    /**
     * A method to trigger a custom-named customer picker input parameter.
     *
     * @param parameterName     The name of the parameter
     * @param attributeName     The attribute which value to return ({@code null} for the entire object)
     * @param defaultFromParent Tells whether to lookup the attribute's value from the parent customer (if any) in case it's <code>null</code> in the customer
     * @return The attribute value
     */
    Object otherCustomer(String parameterName, String attributeName, boolean defaultFromParent) throws PfxClientException {
        return lmApi.callApiUsingFormula(parameterName, attributeName, defaultFromParent);
    }

    /**
     * Equivalent of the standard syntax "CXLookup" function. Retrieves a list of CX entries for the current customer (via context parameter).
     *
     * @param extensionName The name of the extension
     * @param filters       No, one or more {@link #filter(java.lang.String, java.lang.Object)} object/s that narrow down the CX search
     * @return A list of matching CX records
     */
    Object customerExtension(String extensionName, Filter... filters) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{extensionName}, (Object[]) filters);
        Object result = lmApi.callApi(params);
        if (result instanceof List) {
            return MapExtended.convertList((List<Map<String, Object>>) result, TypeCode.CUSTOMEREXTENSION.getShortName(), extensionName);
        }
        return result;
    }

    /**
     * Returns a line item of the given calculable line item collection (a.k.a. clic).
     * Currently, there are three types of clics in total: Quote, Contract, and Rebate Agreement.
     *
     * @param clic   the clic to which the line item belongs. it may either be the clic object as such (its map representation to be exact), or typedId or a uniqueName
     * @param lineId the lineId of the line item
     */
    Object getCalculableLineItem(Object clic, String lineId) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    };

    /**
     * Returns result(s) of the given name(s) (or label(s)) from a line item object (its map representation, to be exact).
     * <p>
     * Note: The line item object will typically be fetched via {@link #getCalculableLineItem(Object, String)}.
     *
     * @return either single result or a map(nameOrLabel,result) of results if there were more names (labels) specified
     */
    Object getCalculableLineItemResult(Object cli, String... resultNamesOrLabels) throws PfxClientException {
        return lmApi.callApiUsingFormula(cli, resultNamesOrLabels);
    }

    /**
     * Returns the full object (header and line items) of e.g. a quote, a rebate agreement or a contract.
     *
     * @return A map specifying the line item collection or null if invalid typedId was passed
     */
    Object getCalculableLineItemCollection(String typedId) throws PfxClientException {
        return lmApi.callApiUsingFormula(typedId);
    }

    /**
     * Creates a revision of CalculableLineItemCollection (Quote/Rebate Agreement/Contract).
     *
     * @return A map specifying the revised line item collection or null if invalid typedId was passed
     */
    Object createNewRevision(String typedId) throws PfxClientException {
        return lmApi.callApiUsingFormula(typedId);
    }

    /**
     * Adds a warning to the formula execution warning log.
     *
     * @param message The warning text message
     */
    void addWarning(String message) {
        ctx.addWarning(message);
    }
    /**
     * Equivalent of the standard syntax "ContextByLabel" function. Retrieves a previous calculation result by formula element label (rather than the element result name).
     * @param label The formula element label
     * @return The calculation result of that element
     */
    Object contextByLabel(String label) throws PfxClientException {
        return lmApi.callApiUsingFormula(label);
    }

    /**
     * Equivalent of the standard syntax "VLookup" function.
     *
     * @param parameterName The name of the price parameter table
     * @return The user selected value. In case of matrix lookups the entire object/row
     */
    Object vLookup(String parameterName) throws PfxClientException {
        return lmApi.callApi(parameterName);
    }

    /**
     * Equivalent of the standard syntax "VLookup" function.
     * <p>
     * Example:
     * <blockquote>
     * <p>
     * <pre>
     *
     * ownShippingCost = api.vLookup("Shipping_Costs", api.product("attribute25").toString())
     *
     * </pre>
     * <p>
     * </blockquote>
     *
     * @param parameterName                  The name of the price parameter table
     * @param attributeNameOrKeyOrRangeValue A predefined key, or range value (hence no user input required), or in case of a matrix lookup the attribute name => user entry required
     * @return The value of the parameter table.
     */
    Object vLookup(String parameterName, Object attributeNameOrKeyOrRangeValue) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeNameOrKeyOrRangeValue);
    }

    /**
     * Equivalent of the standard syntax "VLookup" function. This syntax is only relevant for matrix parameters.
     *<p>
     * Example:
     * <blockquote>
     * <p>
     * <pre>
     *
     * def prevailing = api.vLookup("MixThreshold", "attribute1", "Prevailing")
     *
     * </pre>
     * <p>
     * </blockquote>
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve
     * @param key           The key of the parameter table
     * @return The value of the named attribute for that key
     */
    Object vLookup(String parameterName, String attributeName, String key) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeName, key);
    }

    /**
     * Equivalent of the standard syntax "VLookup" function. This syntax is only relevant for matrix 2-key parameters.
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve
     * @param key1          The key1 of the parameter table
     * @param key2          The key2 of the parameter table
     * @return The value of the named attribute for that key combination
     */
    Object vLookup(String parameterName, String attributeName, String key1, String key2) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeName, key1, key2);
    }

    /**
     * Equivalent of the standard syntax "VLookup" function. This syntax is only relevant for matrix 3-key parameters.
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve
     * @param key1          The key1 of the parameter table
     * @param key2          The key2 of the parameter table
     * @param key3          The key3 of the parameter table
     * @return The value of the named attribute for that key combination
     */
    Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeName, key1, key2, key3);
    }

    /**
     * Equivalent of the standard syntax "VLookup" function. This syntax is only relevant for matrix 4-key parameters.
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve
     * @param key1          The key1 of the parameter table
     * @param key2          The key2 of the parameter table
     * @param key3          The key3 of the parameter table
     * @param key4          The key4 of the parameter table
     * @return The value of the named attribute for that key combination
     */
    Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeName, key1, key2, key3, key4);
    }

    /**
     * Equivalent of the standard syntax "VLookup" function. This syntax is only relevant for matrix 4-key parameters.
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve
     * @param key1          The key1 of the parameter table
     * @param key2          The key2 of the parameter table
     * @param key3          The key3 of the parameter table
     * @param key4          The key4 of the parameter table
     * @param key5          The key5 of the parameter table
     * @return The value of the named attribute for that key combination
     */
    Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4, String key5) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeName, key1, key2, key3, key4, key5);
    }

    /**
     * Equivalent of the standard syntax "VLookup" function. This syntax is only relevant for matrix 4-key parameters.
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve
     * @param key1          The key1 of the parameter table
     * @param key2          The key2 of the parameter table
     * @param key3          The key3 of the parameter table
     * @param key4          The key4 of the parameter table
     * @param key5          The key5 of the parameter table
     * @param key6          The key6 of the parameter table
     * @return The value of the named attribute for that key combination
     */
    Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4, String key5, String key6) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeName, key1, key2, key3, key4, key5, key6);
    }

    /**
     * This syntax can be applied to all MATRIX lookup types.
     * Essentially the keys are passed in a map.
     * The key name can be the technical name ("key1", etc) or the meta-renamed name.
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve
     * @param keys A map with the keys of the row to retrieve. Note: All keys need to be present for the table in question!
     * @return The value of the column {@code attributeName} or full object if attribute name is null
     */
    Object vLookup(String parameterName, String attributeName, Map keys) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeName, keys);
    }

    /**
     * This syntax can be applied to all MATRIX lookup types.
     * Essentially the keys are passed in a map.
     * The key name can be the technical name ("key1", etc) or the meta-renamed name.
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve
     * @param keys A map with the keys of the row to retrieve. Note: All keys need to be present for the table in question!
     * @param defaultValue Default value of the input (relevant for context parameter generation only!)
     * @return The value of the column {@code attributeName} or full object if attribute name is null
     */
    Object vLookup(String parameterName, String attributeName, Map keys, String defaultValue) throws PfxClientException {
        return lmApi.callApi(parameterName, attributeName, keys, defaultValue);
    }


    /**
     * Equivalent of the standard syntax "UserEntry" function. Will trigger a numeric input element.
     *
     * @param entryName The name of the user entry parameter
     * @return The user entered value
     */
    Object userEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Alias of "userEntry", to conform to the general naming structure
     *
     * @param entryName The name of the user entry parameter
     * @return The user entered value
     */
    Object decimalUserEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "IntegerUserEntry" function. It will trigger a numeric input element allowing whole numbers only.
     *
     * @param entryName The name of the user entry parameter
     * @return The user entered value
     */
    Object integerUserEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "StringUserEntry" function. It will trigger a text input element.
     *
     * @param entryName The name of the user entry parameter
     * @return The user entered value
     */
    public Object stringUserEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "StringUserEntry" function. It will trigger a text input element.
     *
     * @param entryName The name of the user entry parameter
     * @param textMask  Optional text mask that is displayed in the UI
     * @return The user entered value
     */
    Object stringUserEntry(String entryName, String textMask) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "BooleanUserEntry" function. It will trigger a Boolean input element.
     *
     * @param entryName The name of the user entry parameter
     * @return The user entered value
     */
    Object booleanUserEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "DateUserEntry" function. It will trigger a date input element.
     * See also: {@link #parseDate(String, String)}
     *
     * @param entryName The name of the user entry parameter
     * @return The user entered value as String in the format yyyy-MM-dd
     */
    Object dateUserEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Creates an input parameter that lets the user enter a date range (start and end date).
     * Rendered as two date picker fields.
     * UNITY UI ONLY!
     *
     * @param entryName The name of the user entry parameter
     * @return String array or list with values entered by the user in the format yyyy-MM-dd
     * @see #parseDate(String, String)
     * @see #timeUserEntry(String)
     * @see #dateTimeUserEntry(String)
     * @see #dateUserEntry(String)
     */
    Object dateRangeUserEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "TimUserEntry" function. It will trigger a time input element.
     * See also: {@link #parseDate(String, String)}
     *
     * @param entryName The name of the user entry parameter
     * @return The user entered value as String in the format HH:mm
     */
    Object timeUserEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "DateTimeUserEntry" function. It will trigger a datetime input element.
     * See also: {@link #parseDate(String, String)}
     *
     * @param entryName The name of the user entry parameter
     * @return The user entered value as String in the format yyyy-MM-dd'T'HH:mm:ss<br><br>
     * (Important note: the format implicitly uses the GMT/UTC time zone and so the returned value may
     * differ from what was entered in on the client. As client time may be in a different time zone.)
     */
    Object dateTimeUserEntry(String entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "ProductGroup" function. It will trigger a product group picker entry.
     *
     * @param entryName The name of the user entry parameter
     * @return The entered product group
     */
    Object productGroupEntry(String... entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Creates an input parameter that lets the user select a group of products and returns the selection as a Map.
     * This is commonly used in PromotionManager and RebateManager.
     * It is rendered as a product group picker widget.
     *
     * @param entryName The name of the user entry parameter
     * @param filterFormulaName A unique name of a formula that will be triggered when the customer picker opens
     * @param filterFormulaParam Additional data for filterFormula
     * @return The entered product group as a Map
     * <pre>
     * [
     *   "productFieldName"  : "attribute3",
     *   "productFieldLabel" : "Business Unit",
     *   "productFieldValue" : "MeatBall"
     * ]
     * </pre>
     * @return
     */
    Object productGroupEntry(String entryName, String filterFormulaName, Object filterFormulaParam) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "CustomerGroup" function. It will trigger a customer group picker entry.
     *
     * @param entryName The name of the user entry parameter
     * @return The entered customer group
     */
    Object customerGroupEntry(String... entryName) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "MultiTierEntry" function. It will trigger a multi tier entry type.
     *
     * @param entryName  The name of the user entry parameter
     * @param arguments Two value hints, sort type and order validation type. Sort type can be one of ASC (default), DESC,
     *               and order validation one of VALIDATE and NO_VALIDATION.
     */
    Object multiTierEntry(String entryName, String... valueHints) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "Option" function. It will trigger a drop down list entry with defined values.
     *
     * @param entryName The name of the user entry parameter
     * @param options   An arbitrary number of parameters that make up the drop down
     * @return The user entered value
     */
    public Object option(String entryName, Object... options) {
        return lmApi.getParameter(entryName);
    }


    /**
     * Equivalent of the standard syntax' "Option" function. It will trigger a drop down list entry with defined values
     * You can add custom (display-only) labels
     *
     * @param entryName The name of the user entry parameter
     * @param options   A list of an arbitrary number of parameters that make up the drop down
     * @param labels	A map (value => label) of labels that should be shown in the UI. If null or empty will behave same as api.option() without label definition
     * @return The user entered value
     */
    public Object option(String entryName, List<Object> options, Map<String, Object> labels) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "FilterBuilderUserEntry" function.
     * <p>
     * It will trigger a FilterBuilder widget allowing to construct a filter object filtering upon the object type represented by the passed
     * <code>typeCode</code> (e.g. to filter upon Products, the <code>typeCode</code> will be <code>P</code>).
     * <p>
     * The filter object will typically be used in other API functions, like {@link #find(String, Filter...)}.
     *
     * @param entryName The name of the user entry parameter
     * @param typeCode  A type code string of the type of the object to filter upon. Supported type codes are P (i.e. Product data) and C (i.e. Customer data), at the moment
     * @return A filter object
     */
    Object filterBuilderUserEntry(String entryName, String typeCode) {
        return lmApi.getParameter(entryName);
    }

    /**
     * Equivalent of the standard syntax "DatamartFilterBuilderUserEntry" function.
     * <p>
     * It will trigger a FilterBuilder widget allowing to construct a filter object filtering upon the specified source.
     * <p>
     * The filter object will typically be used in other API functions, like {@link ISandboxAPI#datamartQuery(Object...)}.
     *
     * @param entryName    The name of the user entry parameter
     * @param source       The name of the source to filter upon.  Resolution of the source as follows
     * <p><ol>
     *                  <li> any datamart, by typedId, sourceName, uniqueName or label  (uniqueName and label for backwards compatibility) </li>
     *                  <li> any FieldCollection, by typedId or sourceName </li>
     * </ol></p>
     * @return A filter object
     */
    @Deprecated
    Object datamartFilterBuilderUserEntry(String entryName, String datamart) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Equivalent of the standard syntax "Pricelist" function.
     *
     * @param listName The name of the pricelist. Target date matching and sku context applies.
     * @return The <code>resultPrice</code> of the pricelist record for the sku
     */
    Object pricelist(String listName) throws PfxClientException {
        return lmApi.callApi(listName);
    }

    /**
     * Equivalent of the standard syntax "Pricelist" function.
     *
     * @param listName      The name of the pricelist. Target date matching and sku context applies.
     * @param attributeName The name of the attribute to retrieve
     * @return The value of the <code>attributeName</code> column of the pricelist record
     */
    Object pricelist(String listName, String attributeName) throws PfxClientException {
        return lmApi.callApi(listName, attributeName);
    }

    /**
     * Equivalent of the standard syntax "Pricelist" function. SKU is taken from parameter and not context.
     *
     * @param listName      The name of the pricelist. Target date matching applies.
     * @param attributeName The name of the attribute to retrieve
     * @param sku           The sku to use
     * @return The value of the <code>attributeName</code> column of the pricelist record
     */
    Object pricelist(String listName, String attributeName, String sku) throws PfxClientException {
        return lmApi.callApi(listName, attributeName, sku);
    }

    /**
     * Equivalent of the standard syntax "Pricelist" function. SKU is taken from parameter and not context.
     *
     * @param listName      The name of the pricelist. Target date matching applies.
     * @param attributeName The name of the attribute to retrieve
     * @param sku           The sku to use
     * @param key2          The key2 to use	 *
     * @return The value of the <code>attributeName</code> column of the matrix pricelist record
     */
    Object pricelist(String listName, String attributeName, String sku, String key2) throws PfxClientException {
        return lmApi.callApi(listName, attributeName, sku, key2);
    }

    /**
     * Equivalent of the standard syntax "Pricelist" function - with the difference that the full item is returned.
     *
     * @param listName The name of the pricelist. Target date matching and sku context applies.
     * @return The entire pricelist item record
     */
    Object pricelistItem(String listName) throws PfxClientException {
        Object result = lmApi.callApi(listName);
        if (result instanceof Map) {
            return new MapExtended((Map<String, Object>) result, TypeCode.PRICELISTITEM.getShortName(), ((Map<String, Object>) result).get(FIELD_PLI_PRICELISTID));
        }
        return result;
    }

    /**
     * Equivalent of the standard syntax "Pricelist" function - with the difference that the full item is returned. SKU is taken from parameter and not context.
     *
     * @param listName The name of the pricelist. Target date matching applies.
     * @param sku      The sku to use
     * @return The entire pricelist item record
     */
    Object pricelistItem(String listName, String sku) throws PfxClientException {
        Object result = lmApi.callApi(listName, sku);
        if (result instanceof Map) {
            return new MapExtended((Map<String, Object>) result, TypeCode.PRICELISTITEM.getShortName(), ((Map<String, Object>) result).get(FIELD_PLI_PRICELISTID));
        }
        return result;
    }

    /**
     * Equivalent of the standard syntax "Pricelist" function - with the difference that the full item is returned. SKU is taken from parameter and not context.
     *
     * @param listName The name of the pricelist. Target date matching applies.
     * @param sku      The sku to use
     * @param key2     The key2 to use
     * @return The entire matrix pricelist item record
     */
    Object pricelistItem(String listName, String sku, String key2) throws PfxClientException {
        Object result = lmApi.callApi(listName, sku, key2);
        if (result instanceof Map) {
            return new MapExtended((Map<String, Object>) result, TypeCode.PRICELISTITEM.getShortName(), ((Map<String, Object>) result).get(FIELD_PLI_PRICELISTID));
        }
        return result;
    }

    /**
     * Looks up the matrix price list item record (XPLI) of the given SKU in a matrix price list of the given name.
     * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
     * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
     *
     * @param listName The name of the price list
     * @param sku      The SKU to use
     * @param key2     The key2 to use
     * @param targetDate The targetDate for teh search
     * @return The entire matrix price list item record (XPLI)
     * @see #pricelist(String, String, String, String)
     */
    Object pricelistItem(String listName, String sku, String key2, Date targetDate) throws PfxClientException {
        Object result = lmApi.callApi(listName, sku, key2, targetDate);
        if (result instanceof Map) {
            return new MapExtended((Map<String, Object>) result, TypeCode.PRICELISTITEM.getShortName(), ((Map<String, Object>) result).get(FIELD_PLI_PRICELISTID));
        }
        return result;
    }

    /**
     * Similar to the standard syntax "Pricelist" function. However it optimizes the subsequent processing by providing the results as map directly.
     *
     * @param listName The name of the pricelist. Target date and sku matching applies.
     */
    Object getPricelistItem(String listName) throws PfxClientException {
        Object result = lmApi.callApi(listName);
        if (result instanceof Map) {
            return new MapExtended((Map<String, Object>) result, TypeCode.PRICELISTITEM.getShortName(), ((Map<String, Object>) result).get(FIELD_PLI_PRICELISTID));
        }
        return result;
    }

    /**
     * Equivalent of the standard syntax "BoM" function. It accesses and retrieves the values from the rolled-up bill of materials for the sku in the context.
     * In this case all quantities are rolled up and summed.
     *
     * @return A list with the summed up quantities
     */
    Object bom() throws PfxClientException {
        return lmApi.callApi();
    }

    /**
     * Equivalent of the standard syntax "BoM" function. It accesses and retrieves the values from the rolled-up bill of materials for the sku in the context.
     * In this case, the summed up quantities are multiplied with a factor that is retrieved by a look up by material name from the provided parameter.
     * An example would be a material-cost table.
     *
     * @param lookupTableName The parameter table name
     * @return A list with the summed up and multiplied quantities
     */
    Object bom(String lookupTableName) throws PfxClientException {
        return lmApi.callApi(lookupTableName);
    }

    /**
     * Equivalent of the standard syntax "BoM" function. It accesses and retrieves the values from the rolled-up bill of materials for the sku in the context.
     * Only BoM records with the specified category are considered.
     * In this case, the summed up quantities are multiplied with a factor that is retrieved by a look up by material name from the provided parameter.
     * An example would be a material-cost table.
     *
     * @param lookupTableName The parameter table name
     * @param categoryName    The category name to filter on
     * @return A list with the summed up and multiplied quantities
     */
    Object bom(String lookupTableName, String categoryName) throws PfxClientException {
        return lmApi.callApi(lookupTableName, categoryName);
    }

    /**
     * Equivalent of the standard syntax "BoM" function. It accesses and retrieves the values from the rolled-up bill of materials for the sku in the context.
     * Only BoM records with the specified category and the material are considered.
     * In this case, the summed up quantities are multiplied with a factor that is retrieved by a look up by material name from the provided parameter.
     * An example would be a material-cost table.
     *
     * @param lookupTableName The parameter table name
     * @param categoryName    The categroy name to filter on
     * @param materialName    The material to filter on
     * @return A list with the summed up and multiplied quantities
     */
    Object bom(String lookupTableName, String categoryName, String materialName) throws PfxClientException {
        return lmApi.callApi(lookupTableName, categoryName, materialName);
    }

    /**
     * Equivalent of the standard syntax "BoMList" function. It rolls up the current sku's BoM and returns the rolled up BoM records (not just the quantities).
     *
     * @return A list of BoM records
     */
    Object bomList() throws PfxClientException {
        return lmApi.callApi();
    }

    /**
     * Returns a value from the current user's master data.
     *
     * @param attributeName The attribute to return
     * @return The attribute value
     */
    Object user(String attributeName) throws PfxClientException {
        return user().get(attributeName);
    }

    /**
     * Returns the entire master data object of the current user.
     *
     * @return The user object
     */
    Map<String, Object> user() throws PfxClientException {
        Map<String, Object> user = lmApi.callApi();
        return new MapExtended(user, TypeCode.USER.name(), (Object) null);
    }

    /**
     * Retrieves an input parameter - without triggering a user input.
     *
     * @param inputName The parameter name
     * @return The parameter value
     */
    Object input(String inputName) {
        return ctx.getParameter(inputName);
    }

    /**
     * Equivalent of the standard "InputMatrix" function. Generates a grid-style input widget with the specified columns.
     * The result will be a list (=the rows) of maps (=the row's columns). The map attribute names are the column names.
     * Make sure the column names are valid JSON identifiers. I.e., spaces are ok but special chars rather not.
     *
     * @param inputMatrixName The name of the parameter
     * @param columnNames     The column names/labels
     * @return A list (=the rows) of maps (=the row's columns)
     */
    Object inputMatrix(String inputMatrixName, String... columnNames) {
        return ctx.getParameter(inputMatrixName);
    }

    /**
     * Creates a filter object that can be used in other API functions. Operator is always "equals", i.e., property = value.
     * <p>
     * Example:
     * <blockquote>
     * <p>
     * <pre>
     *
     * def cas = api.find("CA",api.filter("assignmentId",mplId),api.filter("assignmentType","MPL"))
     *
     * </pre>
     * <p>
     * </blockquote>
     *
     * @param property The property's name to filter on
     * @param value    The specified value
     * @return A filter object
     */
    Object filter(String property, Object value) {
        return new Filter(property, value, Filter.OP_EQUAL);
    }

    /**
     * Constructs a Filter object from the map representation of that filter (from the "wire format").
     * The map representation is the way the filter is described in the JSON requests between UI and server.
     *
     * @param filterMap The filter in form of nested maps
     */
    Filter filterFromMap(Map<String, Object> filterMap) {
        return AdvancedFilterUtil.buildFilterFromAdvancedCriterion(filterMap, null);
    }

    /**
     * Creates a filter object that can be used in other API functions.
     * <p>
     * Note: This method is sort of dynamic. Dynamic in a sense that if the {@code opString} does not actually represent a valid operator (e.g. "="),
     * the passed value will then be treated as lower bound, and the {@code value} as upper bound. In other words, the resulting filter will look like this:
     * {@code Filter.and(Filter.greaterOrEqual(property, opString), Filter.lessThan(property, value))}. If either parameter is {@code null}, its part
     * is skipped.
     * <p>
     * Example:
     * <blockquote>
     * <p>
     * <pre>
     *
     * datamartQuery.where(api.filter("PricingDate",">",targetDate.minus(365).format("yyyy-MM-dd")))
     *
     * </pre>
     * <p>
     * </blockquote>
     *
     * @param property The property's name to filter on
     * @param opString One out of : "=", "<", "<=", ">", ">=", "<>", "~". <em>If different, it gets treated as lower-bound value, then. Please see the remark.</em>
     * @param value    The specified value (<em>or an upper bound</em>)
     * @return A filter object
     */
    Filter filter(String property, Object opString, Object value) {
        Integer op = getFilterOp(opString);

        if (op == null) {
            final Object o1 = opString; //lower boundary
            final Object o2 = value;    //higher boundary
            if (o1 != null || o2 != null) {
                List<Filter> filters = new ArrayList<>();
                if (o1 != null) {
                    filters.add(Filter.greaterOrEqual(property, o1));
                }
                if (o2 != null) {
                    filters.add(Filter.lessThan(property, o2));
                }
                return Filter.and(filters.toArray(new Filter[]{}));
            }
        }

        return new Filter(property, value, op);
    }

    /**
     * Creates a filter to be applied to a search on a related object type, limiting the result to objects of the related type that have a link to the given customer.
     * Example:
     * {@code
     * 		def filter = api.customerToRelatedObjectsFilter("RBA", "104083")	// customerId=104083
     * 		def rbas = api.find("RBA", 0,0, null, filter)
     * 		api.trace("rbas for customer #104083", null, rbas)
     * }
     * @param relatedObjectTypeCode	TypeCode of the related object type.  Possible values are
     * 		<li>'PR':    PriceRecord</li>
     * 		<li>'CT':    Contract</li>
     * 		<li>'CTLI':  ConttractLineItem</li>
     * 		<li>'RBA':   RebateAgreement</li>
     * 		<li>'RBALI': RebateAgreementLineItem</li>
     * @param customerId Customer.id to filter on.
     * @return A filter object
     */
    Filter customerToRelatedObjectsFilter(String relatedObjectTypeCode, String customerId) throws PfxClientException {
        return lmApi.callApi(relatedObjectTypeCode, customerId);
    }

    /**
     * Creates a filter to be applied to a Customer search, limiting the result to customers that are linked to the given related object, the latter defined by its typedId.
     * {@code
     * 		def filter = api.relatedObjectToCustomersFilter("42.RBA")		// rba.id=24
     * 		def customers = api.find("C", 0,0, null, filter)
     * 		api.trace("customers for rba #24", null, customers)
     * }
     * @param relatedObjectTypedId	TypedId of the related object. Its type needs to be one of:
     * 		<li>'PR':    PriceRecord</li>
     * 		<li>'CT':    Contract</li>
     * 		<li>'CTLI':  ConttractLineItem</li>
     * 		<li>'RBA':   RebateAgreement</li>
     * 		<li>'RBALI': RebateAgreementLineItem</li>
     * @return A filter object
     */
    Filter relatedObjectToCustomersFilter(String relatedObjectTypedId) throws PfxClientException {
        return lmApi.callApi(relatedObjectTypedId);
    }

    /**
     * Creates a filter to be applied to a search on a related object type, limiting the result to objects of the related type that have a link to the given product.
     * @param relatedObjectTypeCode	TypeCode of the related object type.  Possible values are:
     * 		<li>'PR':    PriceRecord</li>
     * 		<li>'CT':    Contract</li>
     * 		<li>'CTLI':  ConttractLineItem</li>
     * 		<li>'RBA':   RebateAgreement</li>
     * 		<li>'RBALI': RebateAgreementLineItem</li>
     * @param productId Product.id to filter on.
     * @return A filter object
     */
    Filter productToRelatedObjectsFilter(String relatedObjectTypeCode, String sku) throws PfxClientException {
        return lmApi.callApi(relatedObjectTypeCode, sku);
    }

    /**
     * Creates a filter to be applied to a Product search, limiting the result to products that are linked to the given related object, the latter defined by its typedId.
     * @param relatedObjectTypedId	TypedId of the related object.  Its type needs to be one of:
     * 		<li>'PR':    PriceRecord</li>
     * 		<li>'CT':    Contract</li>
     * 		<li>'CTLI':  ConttractLineItem</li>
     * 		<li>'RBA':   RebateAgreement</li>
     * 		<li>'RBALI': RebateAgreementLineItem</li>
     * @return A filter object
     */
    Filter relatedObjectToProductsFilter(String relatedObjerelatedObjectTypedId) throws PfxClientException {
        return lmApi.callApi(relatedObjerelatedObjectTypedId);
    }

    /**
     * Checks if the current user is member of the given user group.
     *
     * @param userGroupName The group name
     * @param loginUserName The login name of the user
     * @return true or false
     */
    boolean isUserInGroup(String userGroupName, String loginUserName) throws PfxClientException {
        return (Boolean) lmApi.callApi(userGroupName, loginUserName);
    }

    /**
     * Generates a trace message that can be used in the formula test drive. During regular formula execution trace messages are ignored.
     * The parameters specify the output of the trace.
     *
     * @param functionName A free to choose function name
     * @param parameters   Parameters to trace
     * @param result       The result of the trace
     */
    void trace(String functionName, String parameters, Object result) {
        if (ctx.isTraceEnabled()) {
            FormulaTraceMessage trace = new FormulaTraceMessage();
            trace.setFunctionName(functionName);
            trace.setParameters(parameters);
            trace.setResult(result != null ? result.toString() : null);
            ctx.addTraceMessage(trace);
        }
    }

    public void trace(String functionName, Object result) {
        trace(functionName, null, result);
    }

    public void trace(Object result) {
        trace("", null, result);
    }

    /**
     * Retrieves an already generated (i.e. in a previous formula step) input parameter in the form of a <code>ContextParameter</code> object.
     * This can then be further modified as required.
     *
     * @param parameterName The named parameter to retrieve
     * @return The parameter or {@code null}
     */
    ContextParameter getParameter(String parameterName) throws LmDebugUnsupportedException {
/*        if(parameterName == null) return null;

        List<ContextParameter> mParams = ctx.getMissingParameters();
        if(mParams != null) {
            for(ContextParameter cp : mParams) {
                if(parameterName.equals(cp.getName())) return cp;
            }
        }
        return null;*/
        throw new LmDebugUnsupportedException();
    }

    /**
     * In case of list processing (such as a price list or a price grid), this method allows easy access to values from the same list.
     * The result record of the same list with the provided sku is searched and returned.
     * Note: If used within a context that has a secondary key, only the first object is returned!
     *
     * @param sku The sku to look for
     * @return The result record (entire record)
     */
    Map<String, Object> currentContext(String sku) {
        return currentContext(sku,null);
    }

    /**
     * In case of list processing (such as a price list or a price grid), this method allows easy access to values from the same list.
     * The result record of the same list with the provided sku is searched and returned.
     *
     * @param sku  The sku to look for
     * @param key2 The key2 to look for
     * @return The result record (entire record)
     */
    Map<String, Object> currentContext(String sku, String key2) {
//        if(ctx.getCalculationContextCallback() != null)
//            return ctx.getCalculationContextCallback().getCurrentContextResult(sku,key2);
//        else
            return null;
    }

    /**
     * In case of (chained) price lists, this method allows for easy retrieval of records from the previous list.
     * Note: If used within a context that has a secondary key, only the first object is returned!
     *
     * @param sku The sku to look for
     * @return The result record (entire record)
     */
    Map<String, Object> previousContext(String sku) {
        return previousContext(sku, null);
    }

    /**
     * In case of (chained) price lists, this method allows for easy retrieval of records from the previous list.
     *
     * @param sku  The sku to look for
     * @param key2 The key2 to look for
     * @return The result record (entire record)
     */
    Map<String, Object> previousContext(String sku, String key2) {
//        if(ctx.getCalculationContextCallback() != null)
//            return ctx.getCalculationContextCallback().getPreviousContextResult(sku,key2);
//        else
            return null;
    }

    /**
     * The name of the current context. E.g., price list name, quote name, etc.
     *
     * @return The name
     */
    String contextName() {
//        if(ctx.getCalculationContextCallback() != null)
//            return ctx.getCalculationContextCallback().getContextName();
//        else
            return null;
    }

    /**
     * The type (type code string) of the current context.
     */
    String contextType() {
        String type = Utils.convertTypedIdToTypeCode((String) ctx.getParameter(PARAM_FM_ITEM));
        if (type.equals(TypeCode.PRICEGRIDITEM.getShortName())) {
            return TypeCode.PRICEGRID.getShortName();
        } else if (type.equals(TypeCode.PRICELISTITEM.getShortName())) {
            return TypeCode.PRICELIST.getShortName();
        }
        return type;
    }

    /**
     * The typedId of the current context
     */
    String contextTypedId() {
        return (String) ctx.getParameter(PARAM_FM_ITEM);
    }

    /**
     * This method serves as an optimized way to get to certain object attributes. The sandbox automatically converts "real" server objects into
     * a map representation at the time the object is accessed. This is for safety (sandboxing) reasons. This process obviously converts
     * all object attributes. If only one specific attribute is of interest, this method shortcuts that as the underlying object is not exposed
     * (and thus does not force the sandbox into conversion).
     *
     * @param object The object whose value to retrieve
     * @param path   The attribute name
     * @return The value of the attribute
     */
    Object getProperty(Object object, String path) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * This method gets object attributes by elementName. It works only for objects that have meta attributes which contain the element name.
     * These are e.g.: PLI, SIMI, PGI, MPLI
     *
     * @param object      The object whose value to retrieve
     * @param elementName The elementName value
     * @return The value of the attribute or {@code null} if invalid object or no matching meta found
     */
    Object getPropertyByElementName(Map<String, Object> object, String elementName) throws PfxClientException {
        return lmApi.callApi(object, elementName);
    }

    /**
     * In case of a list processing operation this method gets the currently worked on record (e.g., price list item or price grid item or quote item).
     *
     * @return The item object as Map
     */
    MapExtended currentItem() throws PfxClientException {
        MapExtended currentItem = ctx.getCurrentItem();
        if (currentItem == null) {
            Map<String, Object> currentItemMap = lmApi.callApi();
            if (currentItemMap == null) {
                return null;
            }
            String typeCodeStr = Utils.convertTypedIdToTypeCode(currentItemMap.get(FIELD_TYPEDID).toString());
            Object identifier = currentItemMap.get(TypeCode.findByShortName(typeCodeStr).getIdentifierFieldNames()[0]);
            currentItem = new MapExtended(currentItemMap, typeCodeStr, identifier);
            ctx.setCurrentItem(currentItem);
        }
        return currentItem;
    }

    /**
     * In case of a list processing operation, this method gets the currently worked on record's property by elementName.
     *<p>
     *Example:
     *<blockquote>
     * <p>
     * <pre>
     *
     * def currentItem = api.currentItem()
     * def currentSku = currentItem != null ? currentItem.attribute20 : api.product("attribute20")
     *
     * </pre>
     * <p>
     * </blockquote>
     *
     * @param elementName
     * @return The item's property value
     */
    Object currentItemByElementName(String elementName) throws PfxClientException {
        return getPropertyByElementName(currentItem(), elementName);
    }

    /**
     * Same as {@link #currentItem} but returns only the specified attribute.
     *
     * @param attributeName The attribute to retrieve
     * @return The value of the attribute
     */
    Object currentItem(String attributeName) throws PfxClientException {
        return currentItem().get(attributeName);
    }

    /**
     * Returns a set of skus which comprise so-called <em>'sku context'</em> of
     * the current formula execution.
     * <p>
     * The form of the <em>sku context</em> differs, e.g., in case of
     * <em>product detail formula</em> executed via <em>price grid details</em>,
     * the resulting set will contain all skus of the left-pane's grid.
     * <p>
     * Note: If <em>sku context</em> isn't supported by the current formula
     * execution, <code>null</code> is returned.
     * <p>
     * Note #2: A maximum of 200 skus is returned per one call. If there are more of them,
     * <code>startRow</code> must be used to get all skus in the set.
     *
     * @param startRow The starting row of the result set
     * @return a set of skus or <code>null</code> if <em>sku context</em> isn't
     * available for the current formula execution
     */
    Set<Object> contextSkuSet(int startRow) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * The {@link #find(String, Filter...)} method and its variations are limitied in how many results
     * are returned in one call. Some method signatures can specify a maxResults parameter, but there is a system-defined absolute maximum.
     * This call returns the currently system-defined absolute maximum value.
     *
     * @return The currently system-defined absolute maximum number of returned rows by api.find
     */
    int getMaxFindResultsLimit() throws PfxClientException {
        return (Integer) lmApi.callApi();
    }

    /**
     * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
     * The filters are AND'ed together. A maximum of 200 items is returned.
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param filters  a list of filter objects that are AND'ed together
     * @return The result list
     */
    List<Map<String, Object>> find(String typeCode, Filter... filters) throws PfxClientException {
        return find(typeCode,0,null, filters);
    }

    /**
     * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
     * The filters are AND'ed together. A maximum of 200 items is returned.
     * The start row parameter can be used to page through more than 200 results.
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param filters  A list of {@link #filter(java.lang.String, java.lang.Object)} objects that are AND'ed together
     * @param startRow The starting row of the result set
     * @return The result list
     */
    List<Map<String, Object>> find(String typeCode, int startRow, Filter... filters) throws PfxClientException {
        return find(typeCode, startRow,null, filters);
    }

    /**
     * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
     * The filters are AND'ed together. A maximum of 200 items is returned.
     * The start row parameter can be used to page through more than 200 results.
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param filters  A list of filter objects that are AND'ed together
     * @param startRow The starting row of the result set
     * @param sortBy   Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @return The result list
     */
    List<Map<String, Object>> find(String typeCode, int startRow, String sortBy, Filter... filters) throws PfxClientException {
        return find(typeCode, startRow,0, sortBy, filters);
    }

    /**
     * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
     * The filters are AND'ed together. The maximum number of rows retreived is determined by the formulaEngine.script.findMaxResults parameter. The maxRows parameter can only set lower values.
     * <p>
     * Example:
     * <blockquote>
     * <p>
     * <pre>
     *
     * def StockPX = api.find("PX", 0, "-attribute2", Filter.equal("name", "DWHStockReport"), Filter.equal("attribute1", "OFF"), Filter.equal("sku", api.product("sku")), Filter.greaterOrEqual("attribute2", date))
     *
     * </pre>
     * <p>
     * </blockquote>
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param filters  A list of filter objects that are AND'ed together
     * @param startRow The starting row of the result set
     * @param maxRows  The maximum number of rows to retreive. If set to 0 it will default to 200
     * @param sortBy   Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @return The result list
     */

    List<Map<String, Object>> find(String typeCode, int startRow, int maxRows, String sortBy, Filter... filters) throws PfxClientException {
        int rowsLimit = maxRows == 0 || maxRows > ctx.getLmLimitRows() ? ctx.getLmLimitRows() : maxRows;
        return find(typeCode, startRow, rowsLimit, sortBy, null, filters);
    }

    /**
     * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
     * The filters are AND'ed together. The maximum number of rows retreived is determined by the formulaEngine.script.findMaxResults parameter.
     * The maxRows parameter can only set lower values than that.
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param filters  A list of filter objects that are AND'ed together
     * @param startRow The starting row of the result set
     * @param maxRows  The maximum number of rows to retreive. If set to 0 it will default to 200
     * @param sortBy   Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @param fields   A list of fields. If null the entire object will be returned
     * @return The result list
     */
    List<Map<String, Object>> find(String typeCode, int startRow, int maxRows, String sortBy, List<String> fields, Filter... filters) throws PfxClientException {
        return find(typeCode, startRow, maxRows, sortBy, fields, false, filters);
    }

    /**
     * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
     * The filters are AND'ed together. The maximum number of rows retreived is determined by the formulaEngine.script.findMaxResults parameter.
     * The maxRows parameter can only set lower values than that.
     *
     * @param typeCode           A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param filters            A list of filter objects that are AND'ed together
     * @param startRow           The starting row of the result set
     * @param maxRows            The maximum number of rows to retreive. If set to 0 it will default to 200
     * @param sortBy             Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @param fields             A list of fields. If null the entire object will be returned
     * @param distinctValuesOnly fetches distinct/all values, use case is fetching distinct single column values
     * @return The result list
     */
    List find(String typeCode, int startRow, int maxRows, String sortBy, List<String> fields, boolean distinctValuesOnly, Filter... filters) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{typeCode, startRow, maxRows, sortBy, fields, distinctValuesOnly}, (Object[]) filters);
        return MapExtended.convertList(lmApi.callApi(params), typeCode, filters);
    }

    /**
     * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
     * The filters are AND'ed together. The maximum number of rows retreived is determined by the formulaEngine.script.findMaxResults parameter.
     * The maxRows parameter can only set lower values than that.
     *
     * @param typeCode           A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param filters            A list of filter objects that are AND'ed together
     * @param startRow           The starting row of the result set
     * @param maxRows            The maximum number of rows to retreive. If set to 0 it will default to 200
     * @param sortBy             Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @param fields             A map of attributes we only want to return in the result, i.e. skip all attributes not contained in the list the search would return otherwise.
     * 							 The key of the map is the field name. The value can be a field operator like SUM, AVG, MIN or MAX
     * @param distinctValuesOnly fetches distinct/all values, use case is fetching distinct single column values
     * @return The result list
     */
    List find(String typeCode, int startRow, int maxRows, String sortBy, Map<String,String> fields, boolean distinctValuesOnly, Filter ... filters) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{typeCode, startRow, maxRows, sortBy, fields, distinctValuesOnly}, (Object[]) filters);
        return MapExtended.convertList(lmApi.callApi(params), typeCode, filters);
    }

    /**
     * Search for all types of objects with a freestyle query, and iterate over the result set.
     * The filters are AND'ed together.
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code><hostname>/pricefx/<partition name>/fetch</code>
     * @param sortBy   Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @param filters  A list of filter objects that are AND'ed together
     * @return The result implements the Iterator interface, allowing for iteration over all records in scope of the query.
     *
     * Sample groovy code:
     *   def results = api.stream("PCOMP", "id")
     *   def n=0
     *   while(results.hasNext()){
     *   	api.logInfo("result $n", results.next())
     *      n++
     *   }
     */
    ResultIterator stream(String typeCode, String sortBy, Filter... filters) throws PfxClientException {
        Iterator it = find(typeCode, 0, ctx.getLmLimitRows(), sortBy, filters).iterator();
        return new ResultIterator(it);
    }

    /**
     * Search for all types of objects with a freestyle query, and iterate over the result set.
     * The filters are AND'ed together.
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code><hostname>/pricefx/<partition name>/fetch</code>
     * @param sortBy   Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @param fields   A list of attributes we only want to return in the result, i.e. skip all attributes not contained in the list the search would return otherwise
     * @param filters  A list of filter objects that are AND'ed together
     * @return The result implements the Iterator interface, allowing for iteration over all records in scope of the query.
     */
    ResultIterator stream(String typeCode, String sortBy, List<String> fields, Filter... filters) throws PfxClientException {
        Iterator it = find(typeCode, 0, ctx.getLmLimitRows(), sortBy, fields, filters).iterator();
        return new ResultIterator(it);
    }

    /**
     * Search for all types of objects with a freestyle query, and iterate over the result set.
     * The filters are AND'ed together.
     *
     * @param typeCode           A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code><hostname>/pricefx/<partition name>/fetch</code>
     * @param sortBy             Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @param fields             A list of attributes we only want to return in the result, i.e. skip all attributes not contained in the list the search would return otherwise
     * @param distinctValuesOnly fetches distinct/all values, use case is fetching distinct single column values
     * @param filters            A list of filter objects that are AND'ed together
     * @return The result implements the Iterator interface, allowing for iteration over all records in scope of the query.
     */
    ResultIterator stream(String typeCode, String sortBy, List<String> fields, boolean distinctValuesOnly, Filter... filters) throws PfxClientException {
        Iterator it = find(typeCode, 0, ctx.getLmLimitRows(), sortBy, fields, distinctValuesOnly, filters).iterator();
        return new ResultIterator(it);
    }

    /**
     * Search for all types of objects with a freestyle query, and iterate over the result set.
     * The filters are AND'ed together.
     *
     * @param typeCode           A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code><hostname>/pricefx/<partition name>/fetch</code>
     * @param sortBy             Sort the results by an attribute. Prefix with - for descending sort order. Can be comma separated list of sort fields
     * @param fields             A map of attributes we only want to return in the result, i.e. skip all attributes not contained in the list the search would return otherwise.
     * 							 The key of the map is the field name. The value can be a field operator like SUM, AVG, MIN or MAX
     * @param distinctValuesOnly fetches distinct/all values, use case is fetching distinct single column values
     * @param filters            A list of filter objects that are AND'ed together
     * @return The result implements the Iterator interface, allowing for iteration over all records in scope of the query.
     */
    ResultIterator stream(String typeCode, String sortBy, Map<String,String> fields, boolean distinctValuesOnly, Filter ... filters) throws PfxClientException {
        Iterator it = find(typeCode, 0, ctx.getLmLimitRows(), sortBy, fields, distinctValuesOnly, filters).iterator();
        return new ResultIterator(it);
    }
    /**
     * Emits an object to be consumed by an external thread. Requires a blocking emitQueue to be set in the FormulaEngineContext, and
     * a consuming thread to take emitted objects from the queue. See also FeederFormulaExecutor.
     *
     * @param obj The object to be emitted/inserted in the emit queue
     * @return True if the object is successfully inserted in the queue, false otherwise.
     */
    boolean emit(Object obj) {
        return false;
    }

    /**
     * Emits the rebate records matching the given search criteria, to be consumed by a PA allocation task (a PA calculation task with a feeder formula calling
     * this method, and a regular formula accepting each emitted RebateRecord as its currentItem context, and distributing one or more RebateRecord accruals
     * over the set of PA rows which have contributed to those accruals in some way (by revenue, bvolume...).
     *
     * @param dateFieldName	The name of the field to be substitued in the RebateRecord's calculationBase (a DMDataSlice object that includes a generic time dimension filter)
     * @param sortBy		The RebateRecord field on which to sort
     * @param filters		The filters to apply to the RebateRecord search
     */
    void emitRebateRecords(String dateFieldName, String sortBy, Filter... filters) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{dateFieldName, sortBy}, (Object[]) filters);
        lmApi.callApi(params);
    }

    /**
     * Finds the PersistedObject with the given typedId, and emits it if it exists. The assumption is that there is a receiving consumer, as in {@link emitRebateRecords}.
     * @param typedId	The typedId of the PersistedObject to emit
     */
    boolean emitPersistedObject(String typedId) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Finds the PersistedObject for the given typeCode and id, and emits it if it exists. The assumption is that there is a receiving consumer, as in {@link emitRebateRecords}.
     * @param typeCode	The typeCode of the PersistedObject to emit
     * @param id		The id of the PersistedObject to emit
     */
    boolean emitPersistedObject(String typeCode, Long id) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Similar to the {@link #find} methods, this method queries an arbitrary object type. It does not return the objects itself but the count.
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param filters  A list of {@link #filter(java.lang.String, java.lang.Object)} objects that are AND'ed together
     * @return The number of matching records
     */
    int count(String typeCode, Filter... filters) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{typeCode}, (Object[]) filters);
        return (Integer) lmApi.callApi(params);
    }

    /**
     * Finds a lookup table / price parameter table based on the table name. It does apply target date and simulation set matching.
     *
     * @param tableName The name of the lookup table
     * @return The lookup table or {@code null}
     */
    Object findLookupTable(String tableName) throws PfxClientException {
        return lmApi.callApi(tableName);
    }

    /**
     * Searches and returns values of a named lookup table. The returned list can optionally be further filtered.
     *
     * @param tableName The name of the lookup table
     * @param filters   Optional extra {@link #filter(java.lang.String, java.lang.Object)}
     * @return The lookup table values or {@code null}
     */
    List findLookupTableValues(String tableName, Filter... filters) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{tableName}, (Object[]) filters);
        return MapExtended.convertList(lmApi.callApi(params), TypeCode.LOOKUPTABLE.getShortName(), tableName);
    }

    /**
     * Finds active workflows for the given approvable. You can get the same information that is displayed in the workflow UI.
     *
     * @param approvableTypeCode type code of approvable. Approvables are for instances quotes, rebate agreements, data change requests etc.
     * @param approvableId       id of the approvable
     * @return Table like structure resembling workflow UI or {@code null} if not found
     */
    Map<String, Object> findWorkflowInfo(String approvableTypeCode, Object approvableId) throws PfxClientException {
        return lmApi.callApi(approvableTypeCode, approvableId);
    }

    /**
     * Finds previous approvable entity state in case some 'DENIED' or 'WITHDRAWN' workflow for the approvable exists
     * @param typedId Approvable entity typed id
     * @param workflowStatuses Only 'DENIED' or 'WITHDRAWN' are supported
     * @return State of the approvable entity when it's previous workflow was DENIED or WITHDRAWN
     */
    Map<String, Object> findPreviousApprovableState(String typedId, String... workflowStatuses) throws LmDebugUnsupportedException {
		throw new LmDebugUnsupportedException();
	}

	/**
     * Finds approvable entity previous workflow history in case some 'DENIED' or 'WITHDRAWN' workflow for the approvable exists
     * @param typedId Approvable entity typed id
     * @param workflowStatuses Only 'DENIED' or 'WITHDRAWN' are supported
     * @return Workflow history, see net.pricefx.workflowengine.WorkflowInfo
     */
    Object findPreviousWorkflowInfo(String typedId, String... workflowStatuses) throws LmDebugUnsupportedException {
    	throw new LmDebugUnsupportedException();
	}

    /**
     * Finds all assignement records for the given customerId.
     * Also matches by CustomerGroup definitions.
     *
     * @param customerId The customer to search an assignment for
     * @return A list of assignment records
     */
	List<Map<String, Object>> findCustomerAssignments(String customerId) throws PfxClientException {
		return lmApi.callApi(customerId);
	}

	/**
     * Returns a list of pricelists that match the targetDate (i.e., list's targetDate <= targetDate and expiryDate > targetDate).
     *
     * @param targetDate The targetDate to use
     * @param listName   Filter on list name. If {@code null} the list name is not included in search
     * @return The list of matching pricelists
     */
    List<Map<String, Object>> findPricelists(Date targetDate, String listName) throws PfxClientException {
        return lmApi.callApi(targetDate, listName);
    }

    /**
     * Returns a list of simulations that match the targetDate (i.e., list's targetDate <= targetDate).
     *
     * @param targetDate     The targetDate to use
     * @param simulationName Filter on simulation name. If {@code null} the simulation name is not included in search
     * @return The list of matching simulations
     */
    List<Map<String, Object>> findSimulations(Date targetDate, String simulationName) throws PfxClientException {
        return lmApi.callApi(targetDate, simulationName);
    }

    /**
     * Returns a list of price grids that match the targetDate (i.e., price grid's targetDate <= targetDate).
     *
     * @param targetDate    The targetDate to use
     * @param priceGridName Filter on price grid name. If {@code null} the price grid name is not included in search
     * @return The list of matching price grids
     */
    List<Map<String, Object>> findPriceGrids(Date targetDate, String priceGridName) throws PfxClientException {
        return lmApi.callApi(targetDate, priceGridName);
    }

    /**
     * Returns a list of calculated field sets that match the targetDate.
     *
     * @param targetDate Target date for the cfs
     * @param cfsName    Filter on cfs name. If {@code null} the cfs name is not included in search
     * @return The list of matching calculated field sets
     */
    List<Map<String, Object>> findCalculatedFieldSets(Date targetDate, String cfsName) throws PfxClientException {
        return lmApi.callApi(targetDate, cfsName);
    }

    /**
     * Returns a list of manual pricelists that match the targetDate (i.e., list's validAfter <= targetDate and expiryDate > targetDate).
     *
     * @param targetDate The targetDate to use
     * @param listName   Filter on list name. If {@code null} the list name is not included in search
     * @return The list of matching pricelists
     */
    List<Map<String, Object>> findManualPricelists(Date targetDate, String listName) throws PfxClientException {
        return lmApi.callApi(targetDate, listName);
    }

    /**
     * Returns a list of summary items. One item is returned for every matching (=same sku) PLI or XPLI that is in an approved list.
     *
     * @param sku The sku to filter on
     * @return A list of summary items
     */
    List<Map<String, Object>> findApprovedPricelistItems(String sku) throws PfxClientException {
        return lmApi.callApi(sku);
    }

    /**
     * Retrieves a list of customers that match the given CustomerGroup (returns first 200 entries).
     *
     * @param group The group to match
     * @return A list of customers
     */
    List<Map<String, Object>> findCustomersInGroup(Object group) throws PfxClientException {
        return lmApi.callApi(group);
    }

    /**
     * Retrieves a list of customers that match the given CustomerGroup with custom paging and sorting.
     *
     * @param group
     * @param startRow The starting row number (only max 200 rows returned at once)
     * @param sortBy   A sort key. {@code null} if no sorting is required
     * @return A list of customers
     */
    List<Map<String, Object>> findCustomersInGroup(Object group, int startRow, String sortBy) throws PfxClientException {
        return lmApi.callApi(group, startRow, sortBy);
    }

    /**
     * Returns a user entered value that overrides the calculated result of a formula element.
     * <p>
     * Note that the manualOverride works as a flag - when it is set for an element, the final
     * value of that element is taken from the manually entered value and the calculation of that
     * element does not run at all. If this flag is not set, the final value is the calculated result.
     *
     * @param elementName The formula/result element that is overridden
     * @return The overridden value or {@code null} if not overridden
     */
    Object getManualOverride(String elementName) {
        return ctx.getManualResultOverrideValue(elementName);
    }

    /**
     * Removes a user entered value that overrides the calculated result of a formula element.
     * <p>
     * If you have a condition to remove the manual override, it is recommended to put this condition
     * in a separate formula element which is always executed (not in the same one which has the manual
     * override allowed; the reason is that once the user enters a value manually, the element calculation
     * does run at all, so the condition would not be evaluated).
     *
     * @param elementName The formula/result element that is overridden
     */
    void removeManualOverride(String elementName) {
        ctx.removeManualResultOverride(elementName);
    }

    /**
     * Aborts the current item's calculation. Subsequent formula elements are not executed.
     */
    void abortCalculation() {
        ctx.abortCalculation(false);
    }

    /**
     * Aborts the current item calculation. The subsequent logic elements are not executed and
     * the calculation engine continues with calculation of the next item.
     * This function only sets the abortCalculation flag, so the element where abortCalculation is called
     * still returns a value (and is the last element returning a value).
     * This function can be called together with {@link #isSyntaxCheck()} after the parameters are gathered.
     *
     * <p>
     * @param keepOld Should the old outputs be retained? Only applicable for quote and contract calculation!
     */
    void abortCalculation(boolean keepOld) {
        abortCalculation();
    }

    /**
     * Throws a calculation exception with the specified message and aborts the calculation.
     *
     * @param errorMessage The error message that is displayed
     * @throws XExpression
     */
    void throwException(String errorMessage) throws XExpression {
        throw new XExpression(errorMessage);
    }

    /**
     * Marks the specfied item dirty. I.e., second calculation run will be triggered and this item is included in the 2nd pass.
     * If used in a MatrixPricelist/MatrixSimulation (i.e., objects that have a secondary key), all items with that sku are marked as dirty.
     *
     * @param sku The item to mark dirty
     * @throws XExpression
     */
    void markItemDirty(String sku) {
    }

    /**
     * Marks the specfied item dirty. I.e., second calculation run will be triggered and this item is included in the 2nd pass.
     *
     * @param sku
     * @param key2
     */
    void markItemDirty(String sku, String key2) {
    }

    /**
     * Marks the current item dirty. I.e., second calculation run will be triggered and this item is included in the 2nd pass.
     *
     * @throws XExpression
     */
    void markItemDirty() {
    }

    /**
     * Removes a "dirty marker" from the current item.
     *
     * @throws XExpression
     */
    void markItemClean() {
    }

    /**
     * Marks the specfied item clean.
     *
     * @param sku
     * @param key2
     */
    void markItemClean(String sku, String key2) {
    }

    /**
     * Marks the specfied item clean.
     *
     * @param sku
     */
    void markItemClean(String sku) {
    }

    /**
     * A method to determine if the calculation is in the 2nd pass calculation.
     * It is useful for flow control.
     * Note: This does not necessarily mean it is the second iteration (i.e., the first "dirty" iteration), but it is *a* dirty iteration.
     * This is due to the standard default of only one possible dirty run.
     *
     * @return true if is in 2nd pass
     * @throws XExpression
     */
    boolean isSecondPass() {
        return false;
    }

    /**
     * Returns the number of iterations this list has been calculated during this cycle.
     * 0 means: initial calculation (non-dirty),
     * 1 means: first dirty pass
     * and so on
     *
     * @throws XExpression
     */
    int getIterationNumber() {
        return 0;
    }

    /**
     * Formats a string like {@link java.lang.String#format}.
     *
     * @param locale The locale (as string)
     * @param format The format string
     * @param args   The arguments to pass on
     * @return The formatted string
     */
    String formatString(String locale, String format, Object... args) {
        return String.format(new Locale(locale),format,args);
    }

    /**
     * Formats a number into a string. Same semantics as {@link java.text.NumberFormat} and {@link java.text.DecimalFormat#applyPattern}.
     *
     * @param pattern The pattern string
     * @param n       The number to format
     * @return The formatted number as string
     */
    String formatNumber(String pattern, Object n) {
        if (n instanceof String) {
            try {
                n = new BigDecimal((String) n);
            } catch (NumberFormatException nfe) {
            }
        }

        if (n instanceof Number) {
            return new DecimalFormat(pattern).format(n);
        }

        String errorMessage = "API call: formatNumber(pattern, n). Cannot format passed value as a Number. It was: " + (n != null ? n.getClass().getName() : null);
        throw new IllegalArgumentException(errorMessage);
    }

    /**
     * Formats a number into a string. Same semantics as {@link java.text.NumberFormat} and {@link java.text.DecimalFormat#applyPattern}.
     *
     * @param pattern The pattern string
     * @param n       The number to format
     * @param locale  The locae to use for formatting
     * @return The formatted number as string
     */
    String formatNumber(String locale, String pattern, Object n) {
        if (n instanceof String) {
            try {
                n = new BigDecimal((String) n);
            } catch (NumberFormatException nfe) {
            }
        }

        if (n instanceof Number) {
            DecimalFormat decimalFormat = (DecimalFormat) NumberFormat.getNumberInstance(new Locale(locale));
            decimalFormat.applyPattern(pattern);
            return decimalFormat.format(n);
        }

        String errorMessage = "API call: formatNumber(locale, pattern, n). Cannot format passed value as a Number. It was: " + (n != null ? n.getClass().getName() : null);
        throw new IllegalArgumentException(errorMessage);
    }

    /**
     * A utility method to decode a JSON string into a Map representation.
     *
     * @param json The input JSON string
     * @return The parsed JSON as Map
     */
    Object jsonDecode(String json) {
        if (json == null) {
            return new HashMap();
        }
        try {
            return Util.getObjectMapper().readValue(json, new TypeReference<HashMap<String, Object>>() {});
        } catch (Exception ex) {
            String msg = "deserialization error: " + ex.getMessage() + "\ninput: " + json;
            throw new RuntimeException(msg, ex);
        }
    }

    /**
     * A utility method to decode a JSON string into a List representation.
     *
     * @param json The input JSON string
     * @return The parsed JSON as List
     */
    List jsonDecodeList(String json) {
        if (json == null) {
            return new ArrayList();
        }
        try {
            return (List) Util.getObjectMapper().readValue(json, new TypeReference<List<Object>>() {});
        } catch (Exception ex) {
            String msg = "deserialization error: " + ex.getMessage() + "\ninput: " + json;
            throw new RuntimeException(msg, ex);
        }
    }

    /**
     * A utility method to encode Object to JSON string.
     *
     * @param input The input object, usually map
     * @return Encoded object
     */
    String jsonEncode(Object input) {
        if (input == null) {
            return null;
        }
        try {
            return Util.getObjectMapper().writeValueAsString(input);
        }
        catch (Exception e) {
            String msg = "serialization of " + input + " failed: " + e.getMessage();
            throw new RuntimeException(msg, e);
        }
    }

    /**
     * Marks the current formula element with a critical alert and the given message.
     *
     * @param msg The alert message
     */
    void criticalAlert(String msg) {
        ctx.setAlertType("CRITICAL");
        ctx.setAlertMessage(msg);
    }

    /**
     * Marks the current formula element with a red alert and the given message.
     *
     * @param msg The alert message
     */
    void redAlert(String msg) {
        ctx.setAlertType("RED");
        ctx.setAlertMessage(msg);
    }

    /**
     * Marks the current formula element with a yellow alert and the given message.
     *
     * @param msg The alert message
     */
    void yellowAlert(String msg) {
        ctx.setAlertType("YELLOW");
        ctx.setAlertMessage(msg);
    }

    /**
     * Marks the current formula element with the given alert message.
     * If one was already set by means of a yellowAlert, redAlert or criticalAlert call, then that message is replaced.
     * When not raising an alert through any of these calls, but still setting a message with this method,
     * then the message will be displayed wherever calculation results are shown with alert messages, but withoud any yellow or red background coloring.
     *
     * @param msg The alert message
     */
    void setAlertMessage(String msg) {
        ctx.setAlertMessage(msg);
    }


    /**
     * Calls {@link #buildFlexChart(String, Object) buildFlexChart(null,
     * definition)}
     */
    ResultFlexChart buildFlexChart(Object definition) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates a {@link FlexChart} from the passed <code>definition</code> by
     * merging it with <code>options</code> defined within the
     * <code>baseTemplateToUse</code> template.
     * <p>
     * Although it may change in future releases, for now, a
     * <code>FlexChart</code> is nothing else than a <code>Highcharts</code>
     * chart. Therefore the passed <code>definition</code> (also applies to the
     * <code>baseTemplateToUse</code> template) must denote
     * <code>Highcharts options</code> - please consult <code>Highcharts</code>
     * documentation for more details on this.
     * <p>
     * Since <code>Highcharts options</code> are normally specified by a
     * <code>JavaScript object literal</code>, the passed
     * <code>definition</code> must be either a <code>JSON</code> string
     * representing such <code>object literal</code>, or any <code>object</code>
     * which can be mapped to such <code>JSON</code> representation. Therefore,
     * typically the <code>definiton</code> will be passed as a
     * <code>map of maps</code>. Example:
     * <p>
     * <blockquote>
     * <p>
     * <pre>
     * def definition = [
     *   chart: [
     *     type: "bar"
     *   ],
     *   title: [
     *     text: "A common FlexChart"
     *   ],
     *   series: [
     *    [
     *      data: [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21]
     *    ]
     *   ]
     * ]
     *
     * api.buildFlexChart(definition)
     * </pre>
     * <p>
     * </blockquote>
     * <p>
     * Please note that not all <code>Highcharts options</code> are supported!
     * Moreover, usage of some options is prohibited deliberately. These are
     * currently <code>global</code>, <code>events</code> and every
     * <code>JavaScript function</code>-based options.
     * <p>
     * <p>
     * Note: <code>FlexChart</code> templates are defined by administrator in
     * the '<code>Configuration</code>' section.
     * <p>
     *
     * @param baseTemplateToUse the name of the template to use as a basis for merging with
     *                          the passed <code>definition</code>; if {@code null}, the
     *                          predefined <em>Default template</em> gets used
     * @param definition        the definition of the <code>FlexChart</code>; currently, this
     *                          will typically be a string or map of maps containing
     *                          <code>Highcharts options</code>
     * @return a helper object linking the resulting <code>FlexChart</code>,
     * <code>baseTemplateToUse</code> and type of the result - which
     * will always be 'FLEXCHART', here
     * @see <a
     * href="http://api.highcharts.com/highcharts">http://api.highcharts.com/highcharts</a>
     * @see <a
     * href="http://www.highcharts.com/docs">http://www.highcharts.com/docs</a>
     * @see <a href="http://json.org/">http://json.org</a>
     */
    ResultFlexChart buildFlexChart(String baseTemplateToUse,
                                   Object definition) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
     * This query runs against pricelist(s).
     *
     * @return The query definition object
     */
    ItemSummaryQuery getPricelistSummaryQuery() {
        return new ItemSummaryQuery<Pricelist>("pricelistId");
    }

    /**
     * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
     * This query runs against simulation(s).
     *
     * @return The query definition object
     */
    SimulationItemSummaryQuery getSimulationSummaryQuery() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
     * This query runs against price grid(s).
     *
     * @return The query definition object
     */
    ItemSummaryQuery getPriceGridSummaryQuery() {
        return new ItemSummaryQuery<PriceGrid>("priceGridId");
    }

    /**
     * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
     * This query runs against rebate records.
     *
     * @return The query definition object
     */
    RebateRecordSummaryQuery getRebateRecordSummaryQuery() {
        return new RebateRecordSummaryQuery();
    }

    /**
     * Runs (executes) a summary query defined by the parameterized query object, which was created by one of the getXXXSummaryQuery calls.
     *
     * @param query The query to execute
     * @return The result set of the query
     */
    List<Map<String, Object>> runSummaryQuery(ItemSummaryQuery query) throws PfxClientException {
        return lmApi.callApi(query);
    }

    /**
     * Creates a cofigurator entry object and adds the specified input parameter as the first parameter to the configurator entry.
     * If the parameter had a value in the previous round trip, that value is set automatically.
     *
     * @param type Input type of the parameter. Valid enum values are:
     * 	PRODUCT,	PRODUCTGROUP,	CUSTOMER,	CUSTOMERGROUP,	USERENTRY,	STRINGUSERENTRY, OPTION,
     * 	INTEGERUSERENTRY,	BOOLEANUSERENTRY,	DATEUSERENTRY,	TIMEUSERENTRY,	DATETIMEUSERENTRY,	MULTITIERENTRY,
     * 	INPUTMATRIX, OPTIONS,	MATRIXLOOKUP,	LOOKUP,	FILTERBUILDER,	DMFILTERBUILDER,	DMFIELD,	DMDIMFILTER,
     * 	CONFIGURATOR,	HIDDEN
     * @param name name of the parameter
     * @return The configurator entry with the input parameter added
     */
    Object createConfiguratorEntry(InputType type, String name) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates a cofigurator entry object and adds the specified input parameter as the first parameter to the configurator entry.
     * If the parameter had a value in the previous round trip, that value is set automatically.
     *
     * @param type Input type of the parameter. Valid enum values are:
     * 	PRODUCT,	PRODUCTGROUP,	CUSTOMER,	CUSTOMERGROUP,	USERENTRY,	STRINGUSERENTRY, OPTION,
     * 	INTEGERUSERENTRY,	BOOLEANUSERENTRY,	DATEUSERENTRY,	TIMEUSERENTRY,	DATETIMEUSERENTRY,	MULTITIERENTRY,
     * 	INPUTMATRIX, OPTIONS,	MATRIXLOOKUP,	LOOKUP,	FILTERBUILDER,	DMFILTERBUILDER,	DMFIELD,	DMDIMFILTER,
     * 	CONFIGURATOR,	HIDDEN
     * @param name       name of the parameter
     * @param targetDate Target date for the input parameter (not applicable to all input types)
     * @return The configurator entry with the input parameter added
     */
    ConfiguratorEntry createConfiguratorEntry(InputType type, String name, Date targetDate) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates an empty configurator entry object. Must be filled with input parameters by calling createParameter on the object.
     *
     * @return The configurator entry
     */
    ConfiguratorEntry createConfiguratorEntry() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates configurator entry array object which is to be returned from
     * formula element. Allows dynamic setup of configurator entries.
     *
     * @param entries
     *            0..N configurator entries. Supports also list nesting
     * @return The configurator entry array object with defined methods for
     *         further anrichment of array as
     *         {@code ConfiguratorEntryArray#addEntry(ConfiguratorEntry)},
     *         {@code ConfiguratorEntryArray#setEntries(List)} or
     *         {@code ConfiguratorEntryArray#getEntries()}
     */
    ConfiguratorEntryArray createConfiguratorEntryArray(Object ... entries) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Equivalent of the standard syntax "Configurator" function.
     *
     * @param configuratorName The name of the configurator. Similar to other options' names
     * @param formulaName      The name of the configurator's formula which is used to drive the input selections
     * @return The result of the configurator
     */
    Object configurator(String configuratorName, String formulaName) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Equivalent of the standard syntax "Configurator" function.
     *
     * @param configuratorName The name of the configurator. Similar to other options' names
     * @param formulaName      The name of the configurator's formula which is used to drive the input selections
     * @param resultName	   The result that should be returned (instead of the full configurator results map)
     * @return The result of the configurator
     */
    Object configurator(String configuratorName, String formulaName,Object width) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Equivalent of the standard syntax "Configurator" function.
     *
     * @param configuratorName The name of the configurator. Similar to other options' names
     * @param formulaName      The name of the configurator's formula which is used to drive the input selections
     * @param width			   The width of the configurator window. Can be a number (in pixels) or a percentage string.
     * @param height		   The height of the configurator window. Can be a number (in pixels) or a percentage string.
     * @return The result of the configurator
     */
    Object configurator(String configuratorName, String formulaName,Object width, Object height) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Equivalent of the standard syntax' "InlineConfigurator" function.
     *
     * @param configuratorName The name of the configurator. Similar to other options' names
     * @param formulaName      The name of the configurator's formula which is used to drive the input selections
     * @return The result of the configurator
     */
    Object inlineConfigurator(String configuratorName, String formulaName) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }
    /**
     * Parses a date (entered in as string) according to the given pattern.
     * <p>
     * Hint: For parsing datetimes (i.e., dates also including information about the time of day), use the {@link #parseDateTime(String, String)} method instead.
     *
     * @param pattern pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html"
     *                title="DateTimeFormat">DateTimeFormat</a> for pattern format specification)
     * @param date    date as string in format specified by the <code>pattern</code>
     * @return date
     */
    Date parseDate(String pattern, String date) {
        try {
            DateTimeFormatter dtf = DateTimeFormat.forPattern(pattern);
            if(!dtf.isOffsetParsed()) {
                dtf = dtf.withZoneUTC();
            }
            Date result = new Date(dtf.parseMillis(date));
            return result;
        } catch (RuntimeException e) {
        }
        return null;
    }

    /**
     * Parses a datetime (entered in as string) according to the given pattern.
     * <p>
     * Important note: This method will use the UTC time zone for parsing (unless the provided <code>pattern</code> usees a syntax explicitly
     * specifying a time zone). If a different time zone is required, please use the {@link #parseDateTime(String, String, int, int)} method which allows to
     * set a time zone by providing an offset to UTC in hours and minutes.
     *
     * @param pattern  pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html"
     *                 title="DateTimeFormat">DateTimeFormat</a> for pattern format specification)
     * @param datetime datetime as string in format specified by the <code>pattern</code>
     * @return datetime
     */
    DateTime parseDateTime(String pattern, String datetime) {
        try {
            DateTimeFormatter dtf = DateTimeFormat.forPattern(pattern);
            if(!dtf.isOffsetParsed()) {
                dtf = dtf.withZoneUTC();
            }
            DateTime result = dtf.parseDateTime(datetime);
            return result;
        } catch (RuntimeException e) {
        }
        return null;
    }

    /**
     * Parses a datetime (entered in as string) according to a given pattern and time zone.
     * <p>
     * Important note: This method will always use the provided time zone, regardless of the possible time zone set explicitly in <code>datetime</code>.
     *
     * @param pattern         pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html"
     *                        title="DateTimeFormat">DateTimeFormat</a> for pattern format specification)
     * @param datetime        datetime as string in format specified by the <code>pattern</code>
     * @param tzHoursOffset   the offset in hours from UTC, from -23 to +23
     * @param tzMinutesOffset the offset in minutes from UTC, from -59 to +59
     * @return datetime
     */
    DateTime parseDateTime(String pattern, String datetime, int tzHoursOffset, int tzMinutesOffset) {
        try {
            DateTimeFormatter dtf = DateTimeFormat.forPattern(pattern).withZone(DateTimeZone.forOffsetHoursMinutes(tzHoursOffset, tzMinutesOffset));
            DateTime result = dtf.parseDateTime(datetime);
            return result;
        } catch (RuntimeException e) {
        }
        return null;
    }

    /**
     * Creates a cross-list filter expression based on meta data.
     * Example:
     * You'd like to filter accross multiple pricelists on items that have in the column with the elementName "XYZ" the value of "A".
     * Now the elementName XYZ might have been mapped to a different attributeXX column for every list.
     * So a simple attributeXX=A filter does not work.
     * This method does a metadata lookup and creates a proper filter clause based on it for every list.
     * For every list this translates into approx:
     * (attributeXX = A AND listId = <id>)
     * These filters are OR'ed together.
     *
     * @param elementName  The element name (originates from formula element name or MPL integration tag)
     * @param fieldValue   The value (= oprtator applied) of that attribute
     * @param listTypedIds A list of typedId of the "header" objects. E.g. 34.PL. Supported type codes are PL, MPL, PG and SIM
     */
    Filter createElementNameFilter(String elementName, Object fieldValue, String... listTypedIds) throws IOException {
        Object[] params = ArrayUtils.addAll(new Object[]{elementName, fieldValue}, (Object[]) listTypedIds);
        Object result = lmApi.callApi(params);
        return Utils.getObjectMapper().readValue((String) result, Filter.class);
    }

    /**
     * Gets current complete calculation results of the given item (e.g., pricelist item).
     * <p>
     * Note: These are not returned when using other methods (e.g., {@link #pricelistItem(String, String)})
     *
     * @param typedId typedId of the item, e.g. 34.PLI. Supported type codes are PLI, XPLI, SIMI, XSIMI, PGI and RR
     * @return map of individual calculation results with 'Element Name' as key
     */
    Map<String, Object> getItemCompleteCalculationResults(String typedId) throws PfxClientException {
        return lmApi.callApi(typedId);
    }

    /**
     * Issues an HTTP call to external entity. Defaulting to POST request and JSON content type.
     *
     * @param url full url to the endpoint (https or http)
     * @param body body of the request
     * @return Map containing two elements 'responseBody' and 'errorCode'. If content type is JSON and data is returned, responseBody will actually also be a parsed Map and not the raw data.
     */
    Object httpCall(String url, String body) throws PfxClientException {
        return lmApi.callApi(url, body);
    }

    /**
     * Issues an HTTP/S call to Price fx server. Local one or an external one.
     *
     * @param uniqueName            unique name of bound partition. See Configuration/Bound Partitions
     * @param relativeUrl           relative URL to Price f(x) server command. First slash excluded.
     * @param body                  body of the request
     * @param responseBodyAsAString if first param is true, than the body is not converted to json but returned as a string
     * @return Map containing two elements 'responseBody' and 'errorCode'
     */
    Object boundCall(String uniqueName, String relativeUrl, String body, Boolean... responseBodyAsAString) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{uniqueName, relativeUrl, body}, (Object[]) responseBodyAsAString);
        return lmApi.callApi(params);
    }

    /**
     * Returns bound partition list.
     *
     * @return bound partition unique names
     */
    Object getBoundPartitionNames() throws PfxClientException {
        return lmApi.callApi();
    }

    /**
     * Issues an HTTP call to external entity.
     *
     * @param url         - full url to the endpoint (https or http)
     * @param body        body of the request
     * @param methodType  {@code "Get"} or {@code "Post"}
     * @param contentType {@code "JSON"} {@code "XML"} or any full content type for instance {@code "text/plain"}
     * @param properties  map of properties. Following are allowed:
     *                    <ul>
     *                    <li>basicAuthenticationUserName  ->  {@code String} value - user name for basic authentication</li>
     *                    <li>basicAuthenticationPassword -> {@code String} value - user password for basic authentication</li>
     *                    <li>sslTrustAllCertificates {@code boolean} value - Ignores invalid certificates. Use for testing https connections only. Not for production use.</li>
     *                    <li>proxyAuthenticationUserName {@code String} value - user name for proxy (if any)</li>
     *                    <li>proxyAuthenticationPassword {@code String} value - user password for proxy (if any)</li>
     *                    <li>proxyUrl {@code String} value - full url to proxy</li>
     *                    </ul>
     * @return Map containing two elements 'responseBody' and 'statusCode'
     */
    Object httpCall(String url, String body, String methodType, String contentType, @SuppressWarnings("rawtypes") Map... properties) throws PfxClientException {
        Object[] params = ArrayUtils.addAll(new Object[]{url, body, methodType, contentType}, (Object[]) properties);
        return lmApi.callApi(params);
    }

    /**
     * Adds a new object to the collection of objects of the given type.
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param values   key-value pairs representing fields of object we want to add. Note that all significant fields have to be present.
     * @return added object or {@code null} if error occurred
     */
    Object add(String typeCode, Map<String, Object> values) {
        return null;
    }

    /**
     * Adds or updates an object in the collection of objects of the given type.
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param values   key-value pairs representing fields of object we want to add or update. Note that all significant fields have to be present.
     */
    Object addOrUpdate(String typeCode, Map<String, Object> values) {
        return null;
    }

    /**
     * Updates an object in the collection of objects of the given type.
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     *<p>
     *Examples:
     *<blockquote>
     * <p>
     * <pre>
     *
     * api.addOrUpdate("PX", ["name": "Scoring_MM", "sku": sku])
     *
     * api.addOrUpdate("PGI", ["priceGridName" : "LifeCycleMonitor", "sku" : currentItem.sku, "label" : currentItem.label])
     *
     * def large = ["lookupTableId": pp.id, "lookupTableName": pp.uniqueName, "name": "L", "attribute1": percentileHigh, "attribute2": "999999999"]
     * api.addOrUpdate("MLTV", large)
     *
     * </pre>
     * <p>
     * </blockquote>
     *
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param values   key-value pairs representing fields of object we want to update. Note that all significant fields have to be present.
     * @return updated object or {@code null} if error occured
     */
    Object update(String typeCode, Map<String, Object> values) {
        return null;
    }

    /**
     * Deletes an object in the collection of objects of the given type.
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     * <p>
     * Example:
     * <blockquote>
     * <p>
     * <pre>
     *
     * api.delete("PX", ["id": existingRecord, "name": "ActivePrice"])
     *
     * </pre>
     * <p>
     * </blockquote>
     * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling <code>&lt;hostname&gt;/pricefx/&lt;partition name&gt;/fetch</code>
     * @param values   key-value pairs representing significant fields of object we want to delete. Note that all significant fields have to be present, if not the id has to be present
     * @return deleted object or {@code null} if error occured
     */
    Object delete(String typeCode, Map<String, Object> values) {return null;
    }

    /**
     * Sends an email. Only one recipient can be defined.
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     *
     * @param to      email address the email is sent to
     * @param subject email subject
     * @param message email body
     */
    void sendEmail(String to, String subject, String message) {
        //does nothing in debug mode
    }

    /**
     * Sends an event of type "CUSTOM". Other event types are reserved internally.
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     *
     * @param object The object to pass along with the event
     */
    void customEvent(Object object) {
        customEvent(object,null,null);
    }

    /**
     * Sends an event of type "CUSTOM". Other event types are reserved internally.
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     *
     * @param object          The object to pass along with the event
     * @param customEventType A custom string to classify the event type (as the general event type is always CUSTOM)
     */
    void customEvent(Object object, String customEventType) {
        customEvent(object, customEventType, null);
    }

    /**
     * Sends an event of type "CUSTOM". Other event types are reserved internally.
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     *
     * @param object          The object to pass along with the event
     * @param customEventType A custom string to classify the event type (as the general event type is always CUSTOM)
     * @param operation       One of these strings: UPDATE,DELETE,ADD,ITEMCHANGE
     */
    void customEvent(Object object, String customEventType, String operation) {
    }

    /**
     * Updates the header information of a CalculableObject (e.g., a price grid, a price list, etc.).
     * Note: This operation will only work in contexts that allow object modification (CFS,CalculationFlow and direct formula execution via JSON API).
     *
     * @param calcObjTypeCode Type code of object to update
     * @param calcObjName     (Numeric) Id of object to update
     */
    Object updateCalculableObjectHeader(String calcObjTypeCode, String calcObjName) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates an {@link AttributedResult} object to be returned as a result of an output element.
     * <p>
     * <p>
     * Since {@code AttributedResult} is configurable, it's used in cases where it's not sufficient to only return a
     * plain value, but it's also required to adjust formatting, e.g., color, whether the value will display underlined, etc.
     * </p>
     * <p>
     * <p>
     * The usage is simple, instead of returning the output value, create a new {@code AttributedResult} instance like this:
     * {@code api.attributedResult(plain value)}, customize it by calling its various methods, and return it finally.
     * </p>
     * <p>
     * Sample code:<br/>
     * <pre>{@code
     * def cost = api.getElement("Margin_pct")
     * return api.attributedResult(cost)
     * .withBackgroundColor(cost < 0.30 ? "red" : "#0101DF")
     * .withSuffix(cost < 0.30 ? "!!!" : null)
     * .withTextColor(cost < 0.30 ? "white" : null)
     * .withTextDecoration(cost < 0.30 ? "underline" : null)
     * }</pre>
     *
     * @param result calculation result
     * @see AttributedResult AttributedResult - for details on formatting
     */
    AttributedResult attributedResult(Object result) {
        return new AttributedResult(result);
    }

    /**
     * Wraps a result to make sure the original result object appears in the json response.
     * This way also more complex results like lists and maps can be exported as such as a formula result
     *
     * @return A wrapped result object. Should only be used as a return statement of a formula element and not be used otherwise further on
     */
    Object wrap(Object result) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
     */
    ResultMatrix newMatrix() {
        return new ResultMatrix();
    }

    /**
     * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
     * <pre>
     * The individual cells can be styled via the use of:
     * -  linkCell(Object value,String targetPage, String targetPageState)
     * -  imageCell(Object imageUrl)
     * -  styledCell(Object value, String textColor,String bgColor, String weight, String alignment)
     * -  styledCell(Object value, String textColor,String bgColor, String weight)
     * -  styledCell(Object value, String textColor,String bgColor)
     * -  setColumnFormat(String column, FieldFormatType fft)
     *
     * Valid FieldFormatTypes are:
     * NUMERIC, NUMERIC_LONG, MONEY, PERCENT, TEXT, MONEY_EUR, MONEY_USD, MONEY_GBP, MONEY_JPY, MONEY_CHF, MONEY_PLN, DATETIME, DATE, INTEGER
     * </pre>
     * <p>
     * Sample code:
     * <p>
     * <blockquote>
     * <p>
     * <pre>
     *
     * def resultMatrix = api.newMatrix("Margin Status","Customer Id","Name","Net Margin %")
     * def i=0
     * for(entry in rawValues.getEntries()) {
     * 		def row =[:]
     * 		row.put("Customer Id",resultMatrix.linkCell(entry.get("CustomerID"),"customersPage",entry.get("CustomerID")))
     * 		row.put("Name",entry.get("CustomerName"))
     * 		def margin = entry.get("NetMargin%")/100
     * 		row.put("Net Margin %",resultMatrix.styledCell(margin,"#FF0000",null,"bold"))
     * 		if(margin <= 0) row.put("Margin Status",resultMatrix.imageCell("/images/small_trafficlight_red.png"))
     * 		if(margin <0.1 && margin > 0) row.put("Margin Status",resultMatrix.imageCell("/images/small_trafficlight_yellow.png"))
     * 		if(margin >=0.1 ) row.put("Margin Status",resultMatrix.imageCell("/images/small_trafficlight_green.png"))
     * 		resultMatrix.addRow(row)
     * 		i++
     * 		if(i>10) break;
     * }
     * resultMatrix.setColumnFormat("Net Margin %",FieldFormatType.PERCENT)
     * return resultMatrix
     *
     * </pre>
     * <p>
     * </blockquote>
     *
     * @param columns The initial set of columns for the result matrix
     */
    ResultMatrix newMatrix(String... columns) {
        return new ResultMatrix(columns);
    }

    /**
     * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
     *
     * @see newMatrix(String... columns)
     * @param columns The initial set of columns for the result matrix
     *
     * @return
     */
    ResultMatrix newMatrix(Collection<String> columns) {
        return new ResultMatrix(columns);
    }
    /**
     * Creates a new gauge object that can be further customized. Sample:
     * <p>
     * <blockquote>
     * <p>
     * <pre>
     *
     * def g = api.newGauge()
     * g.addSector(5,"#FF0000")
     * g.addSector(25,"#FFFF00")
     * g.addSector(null,"#00FF00")
     * g.setValue(10)
     *
     * </pre>
     * <p>
     * </blockquote>
     */
    ResultGauge newGauge() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates a new dashboard controller object that can be further customized.
     * <blockquote>
     * <p>
     * <pre>
     *
     * def controller = api.newController()
     * controller.addButton(label,targetPage,targetPageState)
     * controller.addHTML(label,someValidHTML)
     *
     * </pre>
     * <p>
     * </blockquote>
     */
    DashboardController newController() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Walks a filter and all its sub filters, visiting each filter in the tree.
     * A pair of closures is used to visit each filter. Each closure can
     * replace the Filter that is visiting. If it does, a new tree will
     * be created for every part of the tree that is affected, thus preserving
     * the original tree. {@code removeNulls} is consulted only in case of
     * (sub)filters which take a list of sub filters.
     * <p>
     * Hence, this method may be used in many cases. For instance, you
     * could use it for mapping the current property set to another one, e.g. to convert
     * a PB filter to a PA one. Somehow like this:
     * <pre>
     * def propertyMappings = ["sku": "ProductID"]
     * api.walkFilter(pbFilter, null, { filter ->
     *  if (filter != null && propertyMappings[filter.property] != null) filter.property = propertyMappings[filter.property]
     *  filter
     * }, false);
     * </pre>
     *
     * @return if any changes have been made, the new Filter; if not, the
     * original Filter.
     */
    Filter walkFilter(Filter filter, Closure<?> visitBefore, Closure<?> visitAfter, boolean removeNulls) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * @param dataLoadLabel Label of DMDataLoad as seen in Data load UI
     * @param dataLoadType  Load Type of DMDataLoad as seen in Data load UI. Namely:
     *                      <ul>
     *                      <li>Customers</li>
     *                      <li>Products</li>
     *                      <li>Flush</li>
     *                      <li>Calendar</li>
     *                      <li>Refresh</li>
     *                      <li>Calculation</li>
     *                      <li>Truncate</li>
     *                      <li>Simulation</li>
     *                      </ul>
     * @param target        Target of DMDataLoad as seen in Data load UI
     * @return dataload if and only if there is exactly one dataload matching (fail early strategy)
     */
    Object findDataLoad(String dataLoadLabel, String dataLoadType, String target) throws PfxClientException {
        Map<String, Object> user = lmApi.callApi(dataLoadLabel, dataLoadType, target);
        return new MapExtended(user, TypeCode.DMDATALOAD.name(), (Object) null);
    }

    /**
     * @param rebateRecordLoadLabel Label of rebate record load as seen in Rebate Record UI
     * @return rebateRecordLoad if and only if there is exactly one load matching (fail early strategy)
     */
    Object findRebateRecordLoad(String rebateRecordLoadLabel) throws PfxClientException {
        Object result = lmApi.callApi(rebateRecordLoadLabel);
        return new MapExtended((Map<String, Object>) result, TypeCode.REBATERECORDSETCALCULATION.name(), (Object) null);
    }

    /**
     * Returns the current partition name.
     * @return The partition name
     */
    String currentPartitionName() {
        return partitionName;
    }

    /**
     * Just a wrapper for org.joda.time.DateTimeZone.forId(String tzString) method.
     * See also the documentation there.
     *
     * @param tzString
     * @return
     */
    DateTimeZone getTimeZone(String tzString) {
        return DateTimeZone.forID(tzString);
    }

    /**
     * Generate a random uuid of the specified length. Example: uuid(15) returns
     * "VcydxgltxrVZSTV"
     *
     * @param len
     *            the desired number of characters
     */
    String uuid(int len) throws PfxClientException {
        return uuid(len, 62);
    }

    /**
     * Generates a random uuid of the specified length, and radix. Examples:
     * <ul>
     * <li>uuid(8, 2) returns "01001010" (8 character ID, base=2)
     * <li>uuid(8, 10) returns "47473046" (8 character ID, base=10)
     * <li>uuid(8, 16) returns "098F4D35" (8 character ID, base=16)
     * </ul>
     *
     * @param len
     *            the desired number of characters
     * @param radix
     *            the number of allowable values for each character (must be <=
     *            62)
     */
    String uuid(int len, int radix) throws PfxClientException {
        return lmApi.callApi(len, radix);
    }

    /**
     * Generates a RFC4122, version 4 ID. Example:
     * "92329D39-6F5C-4520-ABFC-AAB64544E172"
     */
    String uuid() throws PfxClientException {
        return lmApi.callApi();
    }

    /**
     * Constructs a new composite key object, which can be used as the key in a map.
     * Equality of two keys is defined by the pair-wise equality of the corresponding key values.
     *
     * @param keys Key values
     * @return MultiKey composed of the passed in, ordered, key values
     */
    Object multiKey(Object... keys) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Lists all roles available in the application.
     * @return Map of roleUniqueName -> roleLabel pairs
     */
    Map<String, String> roles() throws PfxClientException {
        return lmApi.callApi();
    }

    /**
     * Creates entityRef value (usually db id) for the given entity. The value can be stored in attributeXX fields
     * and you can set its meta type to ENTITY REFERENCE to convince the UI to show the data as links.
     *
     * @return String that represents linkable state of the entity
     */
    String entityRef(Object po) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Evaluates an arithmetic expression that is passed in as a string.
     *
     * Supported operations are:
     * <ul>
     *     <li>addition <code>+</code></li>
     *     <li>subtraction <code>-</code></li>
     *     <li>multiplication <code>*</code></li>
     *     <li>division <code>/</code></li>
     *     <li>exponentiation <code>^</code></li>
     *     <li>square root <code>sqrt</code></li>
     * </ul>
     *
     * Operator precedence and associativity rules are respected.
     *
     * <p>Example:
     * <code>evalExpression("((10 - 3^2 + 1) * -sqrt(1*2+3*4)) / 4")</code>
     * </p>
     *
     * @param mathExpression mathematical expression
     * @return The result of the expression as double
     */
    double evalExpression(String mathExpression) throws PfxClientException {
        Object result = lmApi.callFormulaExpression(mathExpression);
        if (result instanceof Number) {
            return ((Number) result).doubleValue();
        }
        throw new PfxClientException("Cannot convert result of evalExpression() to double");
    }

    /**
     * Stores a string value in the shared cache with the given key. Note the "global"
     * nature of the key!
     *
     * @param key
     *            The cache key
     *
     * @param value
     *            The value to store
     */
    public void setSharedCache(String key, String value) {
        sharedCache.put(key, value);
    }

    /**
     * Retrieves a value from the shared cache (and resets the TTL of the key).
     *
     * @param key
     *            The cache key
     *
     * @return The value or null found in the shared cache
     */
    public String getSharedCache(String key) {
        return sharedCache.get(key);
    }

    /**
     * Deletes a key from the shared chache
     *
     * @param key
     *            The cache key
     */
    public Object removeSharedCache(String key) {
        return sharedCache.remove(key);
    }

    public void log(String msg, Object arg1){
        log.debug("[LmSandboxAPI script] {}: {}", msg, arg1);
    }

    public void log(Object msg) {
        log.debug("[LmSandboxAPI script] {}", msg);
    }

    public void logInfo(String msg, Object arg1){
        log.info("[LmSandboxAPI script] {}: {}", msg, arg1);
    }

    public void logInfo(Object msg){
        log.info("[LmSandboxAPI script] {}", msg);
    }

    public void logWarn(String msg, Object arg1){
        log.warn("[LmSandboxAPI script] {}: {}", msg, arg1);
    }

    public void logWarn(Object msg){
        log.warn("[LmSandboxAPI script] {}", msg);
    }

    private Integer getFilterOp(Object op){
        if (op != null && op instanceof String){
            if("=".equals(op)) {
                return Filter.OP_EQUAL;
            }
            else if("<".equals(op)) {
                return Filter.OP_LESS_THAN;
            }
            else if("<=".equals(op) || "=<".equals(op)) {
                return Filter.OP_LESS_OR_EQUAL;
            }
            else if(">".equals(op)) {
                return Filter.OP_GREATER_THAN;
            }
            else if(">=".equals(op) || "=>".equals(op)) {
                return Filter.OP_GREATER_OR_EQUAL;
            }
            else if("<>".equals(op)) {
                return Filter.OP_NOT_EQUAL;
            }
            else if("~".equals(op)) {
                return Filter.OP_ILIKE;
            }
        }
        return null;
    }
    /**
     * Does a standard Base64 encoding of a string
     *
     * @param string
     * 			The base64 encoded string
     * @return
     */
    String base64Encode(String string) {
        return new String(Base64.encodeBase64(string.getBytes(StandardCharsets.UTF_8)));
    }

    /**
     * Decodes a base64 encoded string
     *
     * @param base64String
     * 			The clear text string
     * @return
     */
    String base64Decode(String base64String) {
        return new String(Base64.decodeBase64(base64String));
    }

    /**
     * Returns a list of product reference records that tell which other objects of given type these skus are referenced
     *
     * @param typeCode Must be either "Q" or "RBA" for now
     * @param skus A set of SKU strings that should be searched for
     * @return A list of reference records
     */
    List<?> getProductReferences(String typeCode, Set<String> skus) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Returns the language part of the current request's locale
     *
     * @return The language part of the current request's locale (e.g. "en")
     */
    String getLocale() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Returns the base URL component of the current instance. Example: https://www.pricefx.eu
     *
     * @return The base URL
     */
    String getBaseURL() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    Object newCreationWorkflow() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    Object newCreationWorkflowStep() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Finds the next revision uniqueNames for the given Quotes/Contracts/RebateAgreements.
     * @param typeCode Can be one of: 'Q', 'CT', 'RBA'
     * @param uniqueNames List of uniqueNames
     * @return Map of [uniqueName : List of nextRevUNs]
     */
    Map<String, List<String>> findNextRevUNs(String typeCode, String... uniqueNames) throws PfxClientException {
        return lmApi.callApi(typeCode, uniqueNames);
    }

    /**
     * Gives current item context in claim validation
     *
     * @return Claim Context instance
     */
    Object getClaimContext() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
//        ctx.getClaimContext();
    }

    /**
     * Indicates whether the current calculation is recalculating the entire list or only parts of it.
     * <p>
     * Note: Currently only applicable for price grid calculations.
     *
     * @return True if the entire grid is calculated, false otherwise. {@code null} is returned if method is not applicable
     */
    Boolean isFullListCalculation() throws LmDebugUnsupportedException {
        return false;
    }

    /**
     * Creates a new background calculation job for the specified price grid.
     * Either a full calculation (parameters == null) or a partial calculation (parameter entry "skusToRecalc" as a list of SKU strings).
     * <p>
     * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct logic execution via JSON API).
     * @param priceGridId The ID of the price grid to calculate
     * @param parameters Optional job parameters
     * @return The ID of the created job tracker
     */
    Long triggerPriceGridCalculation(Long priceGridId, Map<String, Object> parameters)  throws PfxClientException {
        return lmApi.callApi(priceGridId, parameters);
    }

    /**
     * Creates a new background calculation job for the specified CFS.
     * <p>
     * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct logic execution via JSON API).
     * @param cfsID The ID of CalculatedFieldSet
     * @return The ID of the created job tracker
     */
    Long triggerCFSCalculation(Long cfsID) throws PfxClientException {
        return lmApi.callApi(cfsID);
    }

    /**
     * Provides a Map of current execution context info variables.
     * <p>
     * Currently supported variables are:
     * <ul>
     *     <li>
     *         <b>commandName</b> Command name (if any) that was used to invoke current logic execution, e.g. "addproducts", "save", "price", "submit", "calculate"
     *     </li>
     * </ul>
     *
     *
     * @return the context map
     */
    Map currentContext() throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Resets specified fields on currentItem object to null. Applies only to price grid items. Fields that can not be found on the item are skipped.
     *
     * @param fields field names that should be reset
     */
    void resetCurrentItemFields(String... fields) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Evaluates product(group) filter logic
     *
     */
    List<Filter> evalProductFilterLogic(String logicName, String logicParam) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Evaluates customer(group) filter logic
     *
     */
    List<Filter> evalCustomerFilterLogic(String logicName, String logicParam) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }


    /**
     * Creates or updates user input attribute on LPG.
     * <p>
     * <b>Note:</b> Input attributes are saved to extension columns (like attribute12 for example) - same as output elements.
     * Please consider this so that you do not accidentally overwrite any existing output element.
     *
     * @param fieldName Name of field to be used for storing values, e.g. attribute12
     * @param elementName Name of the column, e.g. colour
     * @param label Label of the column, e.g. Colour
     * @param priceGridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
     * @return Created or updated attribute
     */
    Object createOrUpdatePricegridInputAttribute(String fieldName, String elementName, String label, Long priceGridId) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Deletes a pricegrid input attribute.
     *
     * @param id Id of the input attribute to delete
     * @return The deleted attribute object or null if no attrbiute was found
     */
    Object removePricegridInputAttribute(Long id) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Creates or updates user input attribute on PL.
     * <p>
     * <b>Note:</b> Input attributes are saved to extension columns (like attribute12 for example) - same as output elements.
     * Please consider this so that you do not accidentally overwrite any existing output element.
     *
     * @param fieldName Name of field to be used for storing values, e.g. attribute12
     * @param elementName Name of the column, e.g. colour
     * @param label Label of the column, e.g. Colour
     * @param priceGridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
     * @return Created or updated attribute
     */
    Object createOrUpdatePricelistInputAttribute(String fieldName, String elementName, String label, Long pricelistId) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Deletes a pricelist input attribute.
     *
     * @param id Id of the input attribute to delete
     * @return The deleted attribute object or null if no attrbiute was found
     */
    Object removePricelistInputAttribute(Long id) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Sets PG calculation output that will be shown in the header section after PG calculation.
     *
     * @param pricegridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
     * @param name name of the calculation output, e.g. sales
     * @param label	label of the calculation output, e.g. Global Sales
     * @param value calculated value, e.g. €2.3M
     * @param message message containing additional information
     * @return saved calculation output
     */
    Object setPricegridCalculationOutput(Long pricegridId, String name, String label, String value, String message) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Removes PG calculation output.
     *
     * @param pricegridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
     * @param name name of the calculation output that will be removed
     */
    void removePricegridCalculationOutput(Long pricegridId, String name) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

    /**
     * Chart definition that will be used to construct and display a Highchart in the header section.
     *
     * @param definition highchart definition as a map
     * @param pricegridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
     * @return saved chart definition
     */
    Object setPricegridCalculationChart(final Map<String, ?> definition, Long pricegridId) throws LmDebugUnsupportedException {
        throw new LmDebugUnsupportedException();
    }

}
