package net.pricefx.server.dto;

import java.io.Serializable;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import com.fasterxml.jackson.annotation.JsonIgnore;
import net.pricefx.common.api.chart.TieredValueSortType;
import net.pricefx.common.api.chart.TieredValueValidationType;


public class TieredValue implements Serializable {

	private static final long serialVersionUID = -1876343670793908996L;

	public static class Tier implements Serializable {
		private static final long serialVersionUID = -4200298405310577077L;
		int n;
		BigDecimal target;
		BigDecimal value;
		String targetRaw;
		String valueRaw;
		
		Tier(int idx){ n=idx; }
		
		public String toString(){ return (n+1) + ". " + target + "=" + value; }
	}

	
	private List<Tier> tiers = new ArrayList<Tier>();
	private boolean valid;
	private TieredValueSortType sortType = TieredValueSortType.ASC;
	private TieredValueValidationType validationType = TieredValueValidationType.VALIDATE;

	@Override
	public TieredValue clone(){
		return fromMap(asMap(), sortType, validationType);
	}
	
	@JsonIgnore
	public Tier get(int idx){
		if (idx>=0 && idx < tiers.size()) return tiers.get(idx);
		else return null;
	}
	
	public List<Tier> all(){
		return tiers;
	}
	
	public Tier getTierForTarget(BigDecimal target){
		if (target != null){
			Tier targetTier = null;
			for (Tier tier : tiers){
				if (tier.target.compareTo(target) > 0) break;
				targetTier = tier;
			}
			return targetTier;
		}
		return null;
	}
	
	// returns a new TieredValue object with the targets multiplied by a given factor 
	public TieredValue multiplyTargets(BigDecimal factor){
		if (factor == null)
			factor = BigDecimal.ZERO;
		
		TieredValue tv = clone();
		for (Tier tier : tv.tiers){
			tier.target = tier.target.multiply(factor);
		}
		return tv;
	}

	// returns a new TieredValue object with the values multiplied by a given factor
	public TieredValue multiplyValues(BigDecimal factor){
		if (factor == null)
			factor = BigDecimal.ZERO;
		
		TieredValue tv = clone();
		for (Tier tier : tv.tiers){
			tier.value = tier.value.multiply(factor);
		}
		return tv;
	}

	public int size(){
		return tiers.size();
	}
	
	public boolean isValid(){
		return valid;
	}

	public String toString(){
		StringBuilder sb = new StringBuilder();
		boolean first = true;
		for (Tier tier : tiers){
			if (!first) sb.append(", ");
			sb.append(tier.toString());
			first = false;
		}
		return sb.toString();
	}

	public Map<String,String> asMap(){
		Map<String,String> map = new LinkedHashMap<String,String>();
		for (Tier tier : tiers){
			if (tier.target != null || tier.value != null){
				map.put(""+tier.target, ""+tier.value); 
			}
		}
		return map;
	}

	private static Set<Map.Entry<BigDecimal, BigDecimal>> getSortedEntries(TieredValueSortType validationType, Map<BigDecimal, BigDecimal> map) {
		return validationType == TieredValueSortType.DESC
				? ((TreeMap)map).descendingMap().entrySet()
				: map.entrySet();
	}

	public static TieredValue fromMap(final Map<String,String> map) {
		return fromMap(map, TieredValueSortType.ASC, TieredValueValidationType.VALIDATE);
	}
	
	public static TieredValue fromMap(final Map<String,String> map, TieredValueSortType sortType, TieredValueValidationType validationType){
		if (map != null){
			TieredValue tieredVal = new TieredValue();
			tieredVal.valid = true;
			tieredVal.sortType = sortType;
			tieredVal.validationType = validationType;

			Map<BigDecimal, BigDecimal> sortedMap = new TreeMap<>();

			for (Map.Entry<String,String> entry : map.entrySet()) {
				try {
					BigDecimal target = new BigDecimal(entry.getKey());
					BigDecimal value = new BigDecimal(entry.getValue());
					
					sortedMap.put(target, value);
				} catch (NumberFormatException e) {
					tieredVal.valid = false;		
				}
			}
			
			int i=0;
			for (Map.Entry<BigDecimal, BigDecimal> entry : getSortedEntries(sortType, sortedMap)) {
				Tier tier = new Tier(i++);
				tier.targetRaw = entry.getKey().toString();
				tier.valueRaw = entry.getValue().toString();
				tier.target = entry.getKey();
				tier.value = entry.getValue();
				tieredVal.tiers.add(tier);
			}

			return tieredVal;
		}
		return null;
	}
	
}
